# Copyright 2019 QuantRocket LLC - All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# To run: python -m unittest discover -s tests/ -p test*.py -t .

import unittest
try:
    from unittest.mock import patch
except ImportError:
    # py27
    from mock import patch
import pandas as pd
import pytz
import numpy as np
from quantrocket import get_prices
from quantrocket.exceptions import ParameterError, MissingData, NoHistoricalData

class GetPricesTestCase(unittest.TestCase):
    """
    Test cases for `quantrocket.get_prices`.
    """

    def test_complain_if_invalid_timezone(self):
        """
        Tests error handling when the timezone is invalid.
        """

        with self.assertRaises(ParameterError) as cm:
            get_prices("my-db", timezone="Timbuktu")

        self.assertIn("invalid timezone: Timbuktu (see `pytz.all_timezones` for choices)", str(cm.exception))

    @patch("quantrocket.price.list_realtime_databases")
    @patch("quantrocket.price.list_history_databases")
    @patch("quantrocket.price.list_bundles")
    @patch("quantrocket.price.get_history_db_config")
    @patch("quantrocket.price.download_history_file")
    @patch("quantrocket.price.download_master_file")
    def test_pass_args_correctly_single_db(self,
                                            mock_download_master_file,
                                            mock_download_history_file,
                                            mock_get_history_db_config,
                                            mock_list_bundles,
                                            mock_list_history_databases,
                                            mock_list_realtime_databases):
        """
        Tests that args are correctly passed to download_history_file,
        download_master_file, and get_db_config, when there is a single db.
        """
        def _mock_get_history_db_config(db):
            return {
                "bar_size": "1 day",
                "universes": ["usa-stk"],
                "vendor": "ibkr",
                "fields": ["Close","Open","High","Low", "Volume"]
            }

        def _mock_download_history_file(code, f, *args, **kwargs):
            prices = pd.DataFrame(
                dict(
                    Sid=[
                        "FI12345",
                        "FI12345",
                        "FI12345",
                        "FI23456",
                        "FI23456",
                        "FI23456"
                        ],
                    Date=[
                        "2018-04-01",
                        "2018-04-02",
                        "2018-04-03",
                        "2018-04-01",
                        "2018-04-02",
                        "2018-04-03"
                        ],
                    Close=[
                        20.10,
                        20.50,
                        19.40,
                        50.5,
                        52.5,
                        51.59,
                    ]))
            prices.to_csv(f, index=False)

        def _mock_list_history_databases():
            return [
                "usa-stk-1d",
                "demo-stk-1min"
            ]

        def _mock_list_realtime_databases():
            return {}

        def _mock_list_bundles():
            return {}

        mock_list_history_databases.side_effect = _mock_list_history_databases
        mock_list_realtime_databases.side_effect = _mock_list_realtime_databases
        mock_list_bundles.side_effect = _mock_list_bundles

        mock_download_history_file.side_effect = _mock_download_history_file
        mock_get_history_db_config.side_effect = _mock_get_history_db_config

        get_prices(
            "usa-stk-1d", start_date="2018-04-01", end_date="2018-04-03",
            universes="usa-stk", sids=["FI12345","FI23456"], fields=["Close"],
            exclude_universes=["usa-stk-pharm"],
            exclude_sids=[99999], cont_fut=False, timezone="America/New_York")

        mock_download_master_file.assert_not_called()

        self.assertEqual(len(mock_get_history_db_config.mock_calls), 1)
        db_config_call = mock_get_history_db_config.mock_calls[0]
        _, args, kwargs = db_config_call
        self.assertEqual(args[0], "usa-stk-1d")

        mock_list_history_databases.assert_called_once_with()
        mock_list_realtime_databases.assert_called_once_with()
        mock_list_bundles.assert_called_once_with()

        history_call = mock_download_history_file.mock_calls[0]
        _, args, kwargs = history_call
        self.assertEqual(args[0], "usa-stk-1d")
        self.assertListEqual(kwargs["sids"], ["FI12345","FI23456"])
        self.assertEqual(kwargs["start_date"], "2018-04-01")
        self.assertEqual(kwargs["end_date"], "2018-04-03")
        self.assertListEqual(kwargs["fields"], ["Close"])
        self.assertEqual(kwargs["universes"], "usa-stk")
        self.assertListEqual(kwargs["exclude_sids"], [99999])
        self.assertListEqual(kwargs["exclude_universes"], ["usa-stk-pharm"])
        self.assertFalse(kwargs["cont_fut"])

    @patch("quantrocket.price.list_realtime_databases")
    @patch("quantrocket.price.list_history_databases")
    @patch("quantrocket.price.list_bundles")
    @patch("quantrocket.price.get_realtime_db_config")
    @patch("quantrocket.price.get_history_db_config")
    @patch("quantrocket.price.download_market_data_file")
    @patch("quantrocket.price.download_history_file")
    def test_pass_args_correctly_multi_db(
                                        self,
                                        mock_download_history_file,
                                        mock_download_market_data_file,
                                        mock_get_history_db_config,
                                        mock_get_realtime_db_config,
                                        mock_list_bundles,
                                        mock_list_history_databases,
                                        mock_list_realtime_databases):
        """
        Tests that args are correctly passed to download_history_file,
        download_master_file, and get_db_config, when there are multiple dbs,
        including 2 history dbs and 1 realtime db.
        """
        def _mock_get_history_db_config(db):
            if db == "usa-stk-1d":
                return {
                    "bar_size": "1 day",
                    "universes": ["usa-stk"],
                    "vendor": "ibkr",
                    "fields": ["Close","Open","High","Low", "Volume"]
                }
            else:
                return {
                    "bar_size": "1 day",
                    "universes": ["japan-stk"],
                    "vendor": "ibkr",
                    "fields": ["Close","Open","High","Low", "Volume"]
                }

        def _mock_get_realtime_db_config(db):
            return {
                "bar_size": "1 day",
                "fields": ["LastClose", "LastOpen", "VolumeClose"],
                "tick_db_code": "demo-stk-taq"
            }

        def _mock_download_history_file(code, f, *args, **kwargs):
            if code == "usa-stk-1d":
                prices = pd.DataFrame(
                    dict(
                        Sid=[
                            "FI12345",
                            "FI12345",
                            "FI12345",
                            "FI23456",
                            "FI23456",
                            "FI23456"
                            ],
                        Date=[
                            "2018-04-01",
                            "2018-04-02",
                            "2018-04-03",
                            "2018-04-01",
                            "2018-04-02",
                            "2018-04-03"
                            ],
                        Close=[
                            20.10,
                            20.50,
                            19.40,
                            50.5,
                            52.5,
                            51.59,
                        ]))
            else:
                prices = pd.DataFrame(
                    dict(
                        Sid=[
                            "FI56789",
                            "FI56789",
                            "FI56789",
                            ],
                        Date=[
                            "2018-04-01",
                            "2018-04-02",
                            "2018-04-03",
                            ],
                        Close=[
                            5900,
                            5920,
                            5950
                        ]))
            prices.to_csv(f, index=False)

        def _mock_download_market_data_file(code, f, *args, **kwargs):
            prices = pd.DataFrame(
                dict(
                    Sid=[
                        "FI12345",
                        "FI12345",
                        "FI12345",
                        "FI23456",
                        "FI23456",
                        "FI23456"
                        ],
                    Date=[
                        "2018-04-01",
                        "2018-04-02",
                        "2018-04-03",
                        "2018-04-01",
                        "2018-04-02",
                        "2018-04-03"
                        ],
                    LastClose=[
                        20.10,
                        20.50,
                        19.40,
                        50.5,
                        52.5,
                        51.59,
                    ]))
            prices.to_csv(f, index=False)

        def _mock_list_history_databases():
            return [
                "usa-stk-1d",
                "demo-stk-1min",
                "japan-stk-1d"
            ]

        def _mock_list_realtime_databases():
            return {
                "demo-stk-taq": ["demo-stk-taq-1d"]
            }

        def _mock_list_bundles():
            return {}

        mock_download_history_file.side_effect = _mock_download_history_file
        mock_download_market_data_file.side_effect = _mock_download_market_data_file

        mock_get_history_db_config.side_effect = _mock_get_history_db_config
        mock_get_realtime_db_config.side_effect = _mock_get_realtime_db_config

        mock_list_bundles.side_effect = _mock_list_bundles
        mock_list_history_databases.side_effect = _mock_list_history_databases
        mock_list_realtime_databases.side_effect = _mock_list_realtime_databases

        get_prices(
            ["usa-stk-1d", "japan-stk-1d", "demo-stk-taq-1d"],
            start_date="2018-04-01", end_date="2018-04-03",
            sids=["FI12345","FI23456","FI56789"], fields=["Close", "LastClose"])

        mock_list_history_databases.assert_called_once_with()
        mock_list_realtime_databases.assert_called_once_with()
        mock_list_bundles.assert_called_once_with()

        self.assertEqual(len(mock_get_history_db_config.mock_calls), 2)
        db_config_call = mock_get_history_db_config.mock_calls[0]
        _, args, kwargs = db_config_call
        db_config_call_arg1 = args[0]
        db_config_call = mock_get_history_db_config.mock_calls[1]
        _, args, kwargs = db_config_call
        db_config_call_arg2 = args[0]
        self.assertSetEqual({db_config_call_arg1, db_config_call_arg2},
                            {"usa-stk-1d", "japan-stk-1d"})

        self.assertEqual(len(mock_download_history_file.mock_calls), 2)
        history_call = mock_download_history_file.mock_calls[0]
        _, args, kwargs = history_call
        self.assertEqual(args[0], "usa-stk-1d")
        self.assertListEqual(kwargs["sids"], ["FI12345","FI23456","FI56789"])
        self.assertEqual(kwargs["start_date"], "2018-04-01")
        self.assertEqual(kwargs["end_date"], "2018-04-03")
        # only supported subset of fields is requested
        self.assertListEqual(kwargs["fields"], ["Close"])

        history_call = mock_download_history_file.mock_calls[1]
        _, args, kwargs = history_call
        self.assertEqual(args[0], "japan-stk-1d")
        self.assertListEqual(kwargs["sids"], ["FI12345","FI23456","FI56789"])
        self.assertEqual(kwargs["start_date"], "2018-04-01")
        self.assertEqual(kwargs["end_date"], "2018-04-03")
        self.assertListEqual(kwargs["fields"], ["Close"])

        self.assertEqual(len(mock_get_realtime_db_config.mock_calls), 1)
        db_config_call = mock_get_realtime_db_config.mock_calls[0]
        _, args, kwargs = db_config_call
        self.assertEqual(args[0], "demo-stk-taq-1d")

        self.assertEqual(len(mock_download_market_data_file.mock_calls), 1)
        realtime_call = mock_download_market_data_file.mock_calls[0]
        _, args, kwargs = realtime_call
        self.assertEqual(args[0], "demo-stk-taq-1d")
        self.assertListEqual(kwargs["sids"], ["FI12345","FI23456","FI56789"])
        self.assertEqual(kwargs["start_date"], "2018-04-01")
        self.assertEqual(kwargs["end_date"], "2018-04-03")
        # only supported subset of fields is requested
        self.assertListEqual(kwargs["fields"], ["LastClose"])

    def test_query_eod_history_db(self):
        """
        Tests maniuplation of a single EOD db.
        """
        def mock_get_history_db_config(db):
            return {
                "bar_size": "1 day",
                "universes": ["usa-stk"],
                "vendor": "ibkr",
                "fields": ["Close","Open","High","Low", "Volume"]
            }

        def mock_download_history_file(code, f, *args, **kwargs):
            prices = pd.DataFrame(
                dict(
                    Sid=[
                        "FI12345",
                        "FI12345",
                        "FI12345",
                        "FI23456",
                        "FI23456",
                        "FI23456",
                        ],
                    Date=[
                        "2018-04-01",
                        "2018-04-02",
                        "2018-04-03",
                        "2018-04-01",
                        "2018-04-02",
                        "2018-04-03"
                        ],
                    Close=[
                        20.10,
                        20.50,
                        19.40,
                        50.5,
                        52.5,
                        51.59,
                        ],
                    Volume=[
                        15000,
                        7800,
                        12400,
                        98000,
                        179000,
                        142500
                    ]
                ))
            prices.to_csv(f, index=False)

        def mock_list_history_databases():
            return [
                "usa-stk-1d",
                "usa-stk-15min",
                "demo-stk-1min"
            ]

        def mock_list_realtime_databases():
            return {"demo-stk-taq": ["demo-stk-taq-1h"]}

        def mock_list_bundles():
            return {"usstock-1min": True}

        with patch('quantrocket.price.list_bundles', new=mock_list_bundles):
            with patch('quantrocket.price.list_realtime_databases', new=mock_list_realtime_databases):
                with patch('quantrocket.price.list_history_databases', new=mock_list_history_databases):
                    with patch('quantrocket.price.get_history_db_config', new=mock_get_history_db_config):
                        with patch('quantrocket.price.download_history_file', new=mock_download_history_file):

                            prices = get_prices("usa-stk-1d", fields=["Close", "Volume"])

        self.assertListEqual(list(prices.index.names), ["Field", "Date"])
        self.assertEqual(prices.columns.name, "Sid")
        dt = prices.index.get_level_values("Date")
        self.assertTrue(isinstance(dt, pd.DatetimeIndex))
        self.assertIsNone(dt.tz) # EOD is tz-naive
        self.assertListEqual(list(prices.columns), ["FI12345","FI23456"])
        self.assertListEqual(
            list(prices.index.get_level_values("Field")),
            ["Close", "Close", "Close", "Volume", "Volume", "Volume"])
        closes = prices.loc["Close"]
        closes = closes.reset_index()
        closes.loc[:, "Date"] = closes.Date.dt.strftime("%Y-%m-%dT%H:%M:%S%z")
        self.assertListEqual(
            closes.to_dict(orient="records"),
            [{'Date': '2018-04-01T00:00:00', "FI12345": 20.1, "FI23456": 50.5},
             {'Date': '2018-04-02T00:00:00', "FI12345": 20.5, "FI23456": 52.5},
             {'Date': '2018-04-03T00:00:00', "FI12345": 19.4, "FI23456": 51.59}]
        )
        volumes = prices.loc["Volume"]
        volumes = volumes.reset_index()
        volumes.loc[:, "Date"] = volumes.Date.dt.strftime("%Y-%m-%dT%H:%M:%S%z")
        self.assertListEqual(
            volumes.to_dict(orient="records"),
            [{'Date': '2018-04-01T00:00:00', "FI12345": 15000, "FI23456": 98000},
             {'Date': '2018-04-02T00:00:00', "FI12345": 7800, "FI23456": 179000},
             {'Date': '2018-04-03T00:00:00', "FI12345": 12400, "FI23456": 142500}]
        )

    def test_query_multiple_eod_history_dbs(self):
        """
        Tests maniuplation of multiple EOD dbs.
        """
        def mock_get_history_db_config(db):
            if db == "usa-stk-1d":
                return {
                    "bar_size": "1 day",
                    "universes": ["usa-stk"],
                    "vendor": "ibkr",
                    "fields": ["Close","Open","High","Low", "Volume"]
                }
            else:
                return {
                    "bar_size": "1 day",
                    "universes": ["japan-stk"],
                    "vendor": "ibkr",
                    "fields": ["Close","Open","High","Low", "Volume"]
                }

        def mock_download_history_file(code, f, *args, **kwargs):
            if code == "usa-stk-1d":
                prices = pd.DataFrame(
                    dict(
                        Sid=[
                            "FI12345",
                            "FI12345",
                            "FI12345",
                            "FI23456",
                            "FI23456",
                            "FI23456",
                            ],
                        Date=[
                            "2018-04-01",
                            "2018-04-02",
                            "2018-04-03",
                            "2018-04-01",
                            "2018-04-02",
                            "2018-04-03"
                            ],
                        Close=[
                            20.10,
                            20.50,
                            19.40,
                            50.5,
                            52.5,
                            51.59,
                            ],
                        Volume=[
                            15000,
                            7800,
                            12400,
                            98000,
                            179000,
                            142500
                        ]
                    ))
            else:
                prices = pd.DataFrame(
                    dict(
                        Sid=[
                            "FI56789",
                            "FI56789",
                            "FI56789",
                            ],
                        Date=[
                            "2018-04-01",
                            "2018-04-02",
                            "2018-04-03",
                            ],
                        Close=[
                            5900,
                            5920,
                            5950],
                        Volume=[
                            18000,
                            17600,
                            5600
                        ]
                    ))
            prices.to_csv(f, index=False)

        def mock_list_history_databases():
            return [
                "usa-stk-1d",
                "japan-stk-1d",
                "usa-stk-15min",
                "demo-stk-1min"
            ]

        def mock_list_realtime_databases():
            return {"demo-stk-taq": ["demo-stk-taq-1h"]}

        def mock_list_bundles():
            return {"usstock-1min": True}

        with patch('quantrocket.price.list_bundles', new=mock_list_bundles):
            with patch('quantrocket.price.list_realtime_databases', new=mock_list_realtime_databases):
                with patch('quantrocket.price.list_history_databases', new=mock_list_history_databases):
                    with patch('quantrocket.price.get_history_db_config', new=mock_get_history_db_config):
                        with patch('quantrocket.price.download_history_file', new=mock_download_history_file):

                            prices = get_prices(["usa-stk-1d", "japan-stk-1d"], fields=["Close", "Volume"])

        self.assertListEqual(list(prices.index.names), ["Field", "Date"])
        self.assertEqual(prices.columns.name, "Sid")
        dt = prices.index.get_level_values("Date")
        self.assertTrue(isinstance(dt, pd.DatetimeIndex))
        self.assertIsNone(dt.tz) # EOD is tz-naive
        self.assertListEqual(list(prices.columns), ["FI12345","FI23456","FI56789"])
        closes = prices.loc["Close"]
        closes = closes.reset_index()
        closes.loc[:, "Date"] = closes.Date.dt.strftime("%Y-%m-%dT%H:%M:%S%z")
        self.assertListEqual(
            closes.to_dict(orient="records"),
            [{'Date': '2018-04-01T00:00:00', "FI12345": 20.1, "FI23456": 50.5, "FI56789": 5900.0},
             {'Date': '2018-04-02T00:00:00', "FI12345": 20.5, "FI23456": 52.5, "FI56789": 5920.0},
             {'Date': '2018-04-03T00:00:00', "FI12345": 19.4, "FI23456": 51.59, "FI56789": 5950.0}]
        )
        volumes = prices.loc["Volume"]
        volumes = volumes.reset_index()
        volumes.loc[:, "Date"] = volumes.Date.dt.strftime("%Y-%m-%dT%H:%M:%S%z")
        self.assertListEqual(
            volumes.to_dict(orient="records"),
            [{'Date': '2018-04-01T00:00:00', "FI12345": 15000, "FI23456": 98000, "FI56789": 18000},
             {'Date': '2018-04-02T00:00:00', "FI12345": 7800, "FI23456": 179000, "FI56789": 17600},
             {'Date': '2018-04-03T00:00:00', "FI12345": 12400, "FI23456": 142500, "FI56789": 5600}]
        )

    def test_query_eod_history_and_realtime_db(self):
        """
        Tests querying of an EOD history db and EOD realtime aggregate db. This test
        is to make sure we can handle combining a history db with dates like "2020-04-05"
        and a 1-d realtime aggregate db with dates like "2020-04-05T00:00:00+00". It also
        tests that dates are made uniform across all fields.
        """
        def mock_get_history_db_config(db):
            return {
                "bar_size": "1 day",
                "universes": ["usa-stk"],
                "vendor": "ibkr",
                "fields": ["Close","Open","High","Low", "Volume"]
            }

        def mock_download_history_file(code, f, *args, **kwargs):

            prices = pd.DataFrame(
                dict(
                    Sid=[
                        "FI12345",
                        "FI12345",
                        "FI12345",
                        "FI23456",
                        "FI23456",
                        "FI23456",
                        ],
                    Date=[
                        "2018-04-01",
                        "2018-04-02",
                        "2018-04-03",
                        "2018-04-01",
                        "2018-04-02",
                        "2018-04-03"
                        ],
                    Close=[
                        20.10,
                        20.50,
                        19.40,
                        50.5,
                        52.5,
                        51.59,
                        ],
                    Volume=[
                        15000,
                        7800,
                        12400,
                        98000,
                        179000,
                        142500
                    ]
                )
            )
            prices.to_csv(f, index=False)

        def mock_download_market_data_file(code, f, *args, **kwargs):
            prices = pd.DataFrame(
                dict(
                    Sid=[
                        "FI12345",
                        "FI12345",
                        "FI23456",
                        "FI23456",
                        ],
                    Date=[
                        "2018-04-01T00:00:00+00",
                        "2018-04-02T00:00:00+00",
                        "2018-04-01T00:00:00+00",
                        "2018-04-02T00:00:00+00",
                        ],
                    LastClose=[
                        30.50,
                        39.40,
                        79.5,
                        79.59,
                        ],
                    LastCount=[
                        305,
                        940,
                        795,
                        959,
                        ]
                ))
            prices.to_csv(f, index=False)

        def mock_list_history_databases():
            return [
                "usa-stk-1d",
                "japan-stk-1d",
                "usa-stk-15min",
                "demo-stk-1min"
            ]

        def mock_list_realtime_databases():
            return {"usa-stk-tick": ["usa-stk-tick-1d"]}

        def mock_get_realtime_db_config(db):
            return {
                "bar_size": "1 day",
                "universes": ["japan-stk"],
                "vendor": "ibkr",
                "fields": ["LastClose","LastOpen","LastHigh","LastLow", "VolumeClose"]
            }

        def mock_list_bundles():
            return {"usstock-1min": True}

        with patch('quantrocket.price.list_bundles', new=mock_list_bundles):
            with patch('quantrocket.price.list_realtime_databases', new=mock_list_realtime_databases):
                with patch('quantrocket.price.list_history_databases', new=mock_list_history_databases):
                    with patch('quantrocket.price.get_history_db_config', new=mock_get_history_db_config):
                        with patch('quantrocket.price.get_realtime_db_config', new=mock_get_realtime_db_config):
                            with patch('quantrocket.price.download_history_file', new=mock_download_history_file):
                                with patch('quantrocket.price.download_market_data_file', new=mock_download_market_data_file):

                                    prices = get_prices(["usa-stk-1d", "usa-stk-tick-1d"], fields=["Close", "LastClose"])

        self.assertListEqual(list(prices.index.names), ["Field", "Date"])
        self.assertEqual(prices.columns.name, "Sid")
        dt = prices.index.get_level_values("Date")
        self.assertTrue(isinstance(dt, pd.DatetimeIndex))
        self.assertIsNone(dt.tz) # EOD is tz-naive
        self.assertListEqual(list(prices.columns), ["FI12345","FI23456"])
        closes = prices.loc["Close"]
        closes = closes.reset_index()
        closes.loc[:, "Date"] = closes.Date.dt.strftime("%Y-%m-%dT%H:%M:%S%z")
        self.assertListEqual(
            closes.to_dict(orient="records"),
            [{'Date': '2018-04-01T00:00:00', "FI12345": 20.1, "FI23456": 50.5},
             {'Date': '2018-04-02T00:00:00', "FI12345": 20.5, "FI23456": 52.5},
             {'Date': '2018-04-03T00:00:00', "FI12345": 19.4, "FI23456": 51.59}]
        )
        last_closes = prices.loc["LastClose"]
        last_closes = last_closes.reset_index()
        last_closes.loc[:, "Date"] = last_closes.Date.dt.strftime("%Y-%m-%dT%H:%M:%S%z")

        self.assertListEqual(
            last_closes.fillna("nan").to_dict(orient="records"),
            [{'Date': '2018-04-01T00:00:00', "FI12345": 30.50, "FI23456": 79.50},
             {'Date': '2018-04-02T00:00:00', "FI12345": 39.40, "FI23456": 79.59},
             {'Date': '2018-04-03T00:00:00', "FI12345": "nan", "FI23456": "nan"}]
        )

    def test_consolidate_intraday_history_and_realtime_distinct_fields(self):
        """
        Tests that when querying a history and real-time database with
        distinct fields and overlapping dates/sids, both fields are
        preserved.
        """
        def mock_get_history_db_config(db):
            return {
                "bar_size": "15 mins",
                "universes": ["usa-stk"],
                "vendor": "ibkr",
                "fields": ["Close","Open","High","Low", "Volume"]
            }

        def mock_get_realtime_db_config(db):
            return {
                "bar_size": "15 min",
                "fields": ["LastClose","LastOpen","LastHigh","LastLow", "VolumeClose"]
            }
        def mock_download_history_file(code, f, *args, **kwargs):
            prices = pd.DataFrame(
                dict(
                    Sid=[
                        "FI12345",
                        "FI12345",
                        "FI12345",
                        "FI12345",
                        "FI23456",
                        "FI23456",
                        "FI23456",
                        "FI23456",
                        ],
                    Date=[
                        "2018-04-01T09:30:00-04:00",
                        "2018-04-01T15:30:00-04:00",
                        "2018-04-02T09:30:00-04:00",
                        "2018-04-02T15:30:00-04:00",
                        "2018-04-01T09:30:00-04:00",
                        "2018-04-01T15:30:00-04:00",
                        "2018-04-02T09:30:00-04:00",
                        "2018-04-02T15:30:00-04:00",
                        ],
                    Close=[
                        20.10,
                        20.50,
                        19.40,
                        18.56,
                        50.5,
                        52.5,
                        51.59,
                        54.23
                        ],
                    Volume=[
                        15000,
                        7800,
                        12400,
                        14500,
                        98000,
                        179000,
                        142500,
                        124000,
                    ]
                ))
            prices.to_csv(f, index=False)

        def mock_download_market_data_file(code, f, *args, **kwargs):
            prices = pd.DataFrame(
                dict(
                    Sid=[
                        "FI12345",
                        "FI12345",
                        "FI23456",
                        "FI23456",
                        ],
                    Date=[
                        # Data is UTC but will become NY
                        "2018-04-01T19:30:00+00",
                        "2018-04-02T19:30:00+00",
                        "2018-04-01T19:30:00+00",
                        "2018-04-02T19:30:00+00",
                        ],
                    LastClose=[
                        30.50,
                        39.40,
                        79.5,
                        79.59,
                        ]
                ))
            prices.to_csv(f, index=False)

        def mock_download_master_file(f, *args, **kwargs):
            securities = pd.DataFrame(dict(Sid=["FI12345","FI23456"],
                                           Timezone=["America/New_York", "America/New_York"]))
            securities.to_csv(f, index=False)
            f.seek(0)

        def mock_list_history_databases():
            return [
                "usa-stk-15min",
                "japan-stk-1d",
                "demo-stk-1min"
            ]

        def mock_list_realtime_databases():
            return {"usa-stk-snapshot": ["usa-stk-snapshot-15min"]}

        def mock_list_bundles():
            return {"usstock-1min": True}

        with patch('quantrocket.price.list_bundles', new=mock_list_bundles):

            with patch('quantrocket.price.list_realtime_databases', new=mock_list_realtime_databases):
                with patch('quantrocket.price.list_history_databases', new=mock_list_history_databases):
                    with patch('quantrocket.price.get_history_db_config', new=mock_get_history_db_config):
                        with patch('quantrocket.price.get_realtime_db_config', new=mock_get_realtime_db_config):
                            with patch('quantrocket.price.download_history_file', new=mock_download_history_file):
                                with patch('quantrocket.price.download_market_data_file', new=mock_download_market_data_file):
                                    with patch("quantrocket.price.download_master_file", new=mock_download_master_file):

                                        prices = get_prices(
                                            ["usa-stk-15min", "usa-stk-snapshot-15min"],
                                            fields=["Close", "LastClose", "Volume"])

        self.assertListEqual(list(prices.index.names), ["Field", "Date", "Time"])
        self.assertEqual(prices.columns.name, "Sid")
        dt = prices.index.get_level_values("Date")
        self.assertTrue(isinstance(dt, pd.DatetimeIndex))
        self.assertIsNone(dt.tz)
        self.assertListEqual(list(prices.columns), ["FI12345","FI23456"])
        closes = prices.loc["Close"]
        closes = closes.reset_index()
        closes.loc[:, "Date"] = closes.Date.dt.strftime("%Y-%m-%dT%H:%M:%S%z")
        self.assertListEqual(
            closes.to_dict(orient="records"),
            [{'Date': '2018-04-01T00:00:00', 'Time': '09:30:00', "FI12345": 20.1, "FI23456": 50.5},
             {'Date': '2018-04-01T00:00:00', 'Time': '15:30:00', "FI12345": 20.5, "FI23456": 52.5},
             {'Date': '2018-04-02T00:00:00', 'Time': '09:30:00', "FI12345": 19.4, "FI23456": 51.59},
             {'Date': '2018-04-02T00:00:00', 'Time': '15:30:00', "FI12345": 18.56, "FI23456": 54.23}]
        )
        volumes = prices.loc["Volume"]
        volumes = volumes.reset_index()
        volumes.loc[:, "Date"] = volumes.Date.dt.strftime("%Y-%m-%dT%H:%M:%S%z")
        self.assertListEqual(
            volumes.to_dict(orient="records"),
            [{'Date': '2018-04-01T00:00:00', 'Time': '09:30:00', "FI12345": 15000.0, "FI23456": 98000.0},
             {'Date': '2018-04-01T00:00:00', 'Time': '15:30:00', "FI12345": 7800.0, "FI23456": 179000.0},
             {'Date': '2018-04-02T00:00:00', 'Time': '09:30:00', "FI12345": 12400.0, "FI23456": 142500.0},
             {'Date': '2018-04-02T00:00:00', 'Time': '15:30:00', "FI12345": 14500.0, "FI23456": 124000.0}]
        )

        last_closes = prices.loc["LastClose"]
        last_closes = last_closes.reset_index()
        last_closes.loc[:, "Date"] = last_closes.Date.dt.strftime("%Y-%m-%dT%H:%M:%S%z")
        self.assertListEqual(
            last_closes.fillna("nan").to_dict(orient="records"),
            # Data was UTC but now NY
            [{'Date': '2018-04-01T00:00:00', 'Time': '09:30:00', "FI12345": "nan", "FI23456": "nan"},
             {'Date': '2018-04-01T00:00:00', 'Time': '15:30:00', "FI12345": 30.5, "FI23456": 79.5},
             {'Date': '2018-04-02T00:00:00', 'Time': '09:30:00', "FI12345": "nan", "FI23456": "nan"},
             {'Date': '2018-04-02T00:00:00', 'Time': '15:30:00', "FI12345": 39.4, "FI23456": 79.59}]
        )

    def test_query_single_realtime_db(self):
        """
        Tests querying a single real-time aggregate database, with no history
        db in the query.
        """
        def mock_get_realtime_db_config(db):
            return {
                "bar_size": "15 min",
                "fields": ["LastClose","LastOpen","LastHigh","LastLow", "VolumeClose"]
            }

        def mock_download_market_data_file(code, f, *args, **kwargs):
            prices = pd.DataFrame(
                dict(
                    Sid=[
                        "FI12345",
                        "FI12345",
                        "FI23456",
                        "FI23456",
                        ],
                    Date=[
                        # Data is UTC but will become Mexico_City
                        "2018-04-01T19:30:00+00",
                        "2018-04-02T19:30:00+00",
                        "2018-04-01T19:30:00+00",
                        "2018-04-02T19:30:00+00",
                        ],
                    LastClose=[
                        30.50,
                        39.40,
                        79.5,
                        79.59,
                        ],
                    LastCount=[
                        305,
                        940,
                        795,
                        959,
                        ]
                ))
            prices.to_csv(f, index=False)

        def mock_download_master_file(f, *args, **kwargs):
            securities = pd.DataFrame(dict(Sid=["FI12345","FI23456"],
                                           Timezone=["America/Mexico_City", "America/Mexico_City"]))
            securities.to_csv(f, index=False)
            f.seek(0)

        def mock_list_history_databases():
            return [
                "usa-stk-15min",
                "japan-stk-1d",
                "demo-stk-1min"
            ]

        def mock_list_realtime_databases():
            return {"mexi-stk-tick": ["mexi-stk-tick-15min"]}

        def mock_list_bundles():
            return {"usstock-1min": True}

        with patch('quantrocket.price.list_bundles', new=mock_list_bundles):
            with patch('quantrocket.price.list_realtime_databases', new=mock_list_realtime_databases):
                with patch('quantrocket.price.list_history_databases', new=mock_list_history_databases):
                    with patch('quantrocket.price.get_realtime_db_config', new=mock_get_realtime_db_config):
                        with patch('quantrocket.price.download_market_data_file', new=mock_download_market_data_file):
                            with patch("quantrocket.price.download_master_file", new=mock_download_master_file):

                                prices = get_prices("mexi-stk-tick-15min")

        self.assertListEqual(list(prices.index.names), ["Field", "Date", "Time"])
        self.assertEqual(prices.columns.name, "Sid")
        dt = prices.index.get_level_values("Date")
        self.assertTrue(isinstance(dt, pd.DatetimeIndex))
        self.assertIsNone(dt.tz)
        self.assertListEqual(list(prices.columns), ["FI12345","FI23456"])

        last_closes = prices.loc["LastClose"]
        last_closes = last_closes.reset_index()
        last_closes.loc[:, "Date"] = last_closes.Date.dt.strftime("%Y-%m-%dT%H:%M:%S%z")
        self.assertListEqual(
            last_closes.fillna("nan").to_dict(orient="records"),
            # Data was UTC but now Mexico_City
            [{'Date': '2018-04-01T00:00:00', 'Time': '14:30:00', "FI12345": 30.5, "FI23456": 79.5},
             {'Date': '2018-04-02T00:00:00', 'Time': '14:30:00', "FI12345": 39.4, "FI23456": 79.59}]
        )

        last_counts = prices.loc["LastCount"]
        last_counts = last_counts.reset_index()
        last_counts.loc[:, "Date"] = last_counts.Date.dt.strftime("%Y-%m-%dT%H:%M:%S%z")
        self.assertListEqual(
            last_counts.fillna("nan").to_dict(orient="records"),
            # Data was UTC but now Mexico_City
            [{'Date': '2018-04-01T00:00:00', 'Time': '14:30:00', "FI12345": 305.0, "FI23456": 795.0},
             {'Date': '2018-04-02T00:00:00', 'Time': '14:30:00', "FI12345": 940.0, "FI23456": 959.0}]
        )

    def test_query_single_zipline_bundle(self):
        """
        Tests querying a single Zipline bundle, with no history
        db or realtime db in the query.
        """
        def mock_download_minute_file(code, f, *args, **kwargs):
            prices = pd.concat(
                dict(
                    Close=pd.DataFrame(
                        dict(
                            FI12345=[
                                48.90,
                                49.40,
                                55.49,
                                56.78
                            ],
                            FI23456=[
                                59.5,
                                59.59,
                                59.34,
                                51.56,
                            ]
                        ),
                        index=[
                            "2018-04-01T09:30:00-04:00",
                            "2018-04-01T15:30:00-04:00",
                            "2018-04-02T09:30:00-04:00",
                            "2018-04-02T15:30:00-04:00",
                            ]),
                    Volume=pd.DataFrame(
                        dict(
                            FI12345=[
                                100,
                                200,
                                300,
                                400
                            ],
                            FI23456=[
                                500,
                                600,
                                700,
                                800,
                            ]
                        ),
                        index=[
                            "2018-04-01T09:30:00-04:00",
                            "2018-04-01T15:30:00-04:00",
                            "2018-04-02T09:30:00-04:00",
                            "2018-04-02T15:30:00-04:00",
                            ])
                    ), names=["Field","Date"]
            )
            prices.to_csv(f)

        def mock_download_master_file(f, *args, **kwargs):
            securities = pd.DataFrame(dict(Sid=["FI12345","FI23456"],
                                           Timezone=["America/New_York", "America/New_York"]))
            securities.to_csv(f, index=False)
            f.seek(0)

        def mock_list_history_databases():
            return [
                "usa-stk-15min",
                "japan-stk-1d",
                "demo-stk-1min"
            ]

        def mock_list_realtime_databases():
            return {"mexi-stk-tick": ["mexi-stk-tick-15min"]}

        def mock_list_bundles():
            return {"usstock-1min": True}

        with patch('quantrocket.price.list_bundles', new=mock_list_bundles):
            with patch('quantrocket.price.list_realtime_databases', new=mock_list_realtime_databases):
                with patch('quantrocket.price.list_history_databases', new=mock_list_history_databases):
                    with patch('quantrocket.price.download_minute_file', new=mock_download_minute_file):
                        with patch("quantrocket.price.download_master_file", new=mock_download_master_file):

                            prices = get_prices("usstock-1min", fields=["Close","Volume"])

        self.assertListEqual(list(prices.index.names), ["Field", "Date", "Time"])
        self.assertEqual(prices.columns.name, "Sid")
        dt = prices.index.get_level_values("Date")
        self.assertTrue(isinstance(dt, pd.DatetimeIndex))
        self.assertIsNone(dt.tz)
        self.assertListEqual(list(prices.columns), ["FI12345","FI23456"])

        closes = prices.loc["Close"]
        closes = closes.reset_index()
        closes.loc[:, "Date"] = closes.Date.dt.strftime("%Y-%m-%dT%H:%M:%S%z")
        self.assertListEqual(
            closes.fillna("nan").to_dict(orient="records"),
            [{'Date': '2018-04-01T00:00:00', 'Time': '09:30:00', 'FI12345': 48.9, 'FI23456': 59.5},
             {'Date': '2018-04-01T00:00:00', 'Time': '15:30:00', 'FI12345': 49.4, 'FI23456': 59.59},
             {'Date': '2018-04-02T00:00:00', 'Time': '09:30:00', 'FI12345': 55.49, 'FI23456': 59.34},
             {'Date': '2018-04-02T00:00:00', 'Time': '15:30:00', 'FI12345': 56.78, 'FI23456': 51.56}]
        )

        volumes = prices.loc["Volume"]
        volumes = volumes.reset_index()
        volumes.loc[:, "Date"] = volumes.Date.dt.strftime("%Y-%m-%dT%H:%M:%S%z")
        self.assertListEqual(
            volumes.fillna("nan").to_dict(orient="records"),
            [{'Date': '2018-04-01T00:00:00', 'Time': '09:30:00', 'FI12345': 100, 'FI23456': 500},
             {'Date': '2018-04-01T00:00:00', 'Time': '15:30:00', 'FI12345': 200, 'FI23456': 600},
             {'Date': '2018-04-02T00:00:00', 'Time': '09:30:00', 'FI12345': 300, 'FI23456': 700},
             {'Date': '2018-04-02T00:00:00', 'Time': '15:30:00', 'FI12345': 400, 'FI23456': 800}]
        )

    @patch("quantrocket.price.download_market_data_file")
    @patch("quantrocket.price.download_history_file")
    @patch("quantrocket.price.download_minute_file")
    def test_apply_times_filter_to_history_vs_realtime_database_vs_zipline_bundle(
        self,
        mock_download_minute_file,
        mock_download_history_file,
        mock_download_market_data_file):

        """
        Tests that the times filter is applied to a history database via the
        history query but is applied to the realtime database after
        converting to the exchange timezone.
        """
        def mock_get_history_db_config(db):
            return {
                "bar_size": "1 mins",
                "universes": ["usa-stk"],
                "vendor": "ibkr",
                "fields": ["Wap","Open","High","Low", "Volume"]
            }

        def mock_get_realtime_db_config(db):
            return {
                "bar_size": "1 min",
                "fields": ["LastClose","LastOpen","LastHigh","LastLow", "VolumeClose"]
            }
        def _mock_download_history_file(code, f, *args, **kwargs):
            prices = pd.DataFrame(
                dict(
                    Sid=[
                        "FI12345",
                        "FI12345",
                        "FI12345",
                        "FI12345",
                        "FI23456",
                        "FI23456",
                        "FI23456",
                        "FI23456",
                        ],
                    Date=[
                        "2018-04-01T09:30:00-04:00",
                        "2018-04-01T15:30:00-04:00",
                        "2018-04-02T09:30:00-04:00",
                        "2018-04-02T15:30:00-04:00",
                        "2018-04-01T09:30:00-04:00",
                        "2018-04-01T15:30:00-04:00",
                        "2018-04-02T09:30:00-04:00",
                        "2018-04-02T15:30:00-04:00",
                        ],
                    Wap=[
                        20.10,
                        20.50,
                        19.40,
                        18.56,
                        50.5,
                        52.5,
                        51.59,
                        54.23
                        ],
                    Volume=[
                        15000,
                        7800,
                        12400,
                        14500,
                        98000,
                        179000,
                        142500,
                        124000,
                    ]
                ))
            prices.to_csv(f, index=False)

        def _mock_download_market_data_file(code, f, *args, **kwargs):
            prices = pd.DataFrame(
                dict(
                    Sid=[
                        "FI12345",
                        "FI12345",
                        "FI12345",
                        "FI12345",
                        "FI23456",
                        "FI23456",
                        "FI23456",
                        "FI23456",
                        ],
                    Date=[
                        # 19:00:00 UTC data is returned (15:00:00 NY) but will be filtered out
                        "2018-04-01T19:00:00+00",
                        "2018-04-01T19:30:00+00",
                        "2018-04-02T19:00:00+00",
                        "2018-04-02T19:30:00+00",
                        "2018-04-01T19:00:00+00",
                        "2018-04-01T19:30:00+00",
                        "2018-04-02T19:00:00+00",
                        "2018-04-02T19:30:00+00",
                        ],
                    LastClose=[
                        30.50,
                        39.40,
                        45.49,
                        46.78,
                        79.5,
                        79.59,
                        89.34,
                        81.56,
                        ]
                ))
            prices.to_csv(f, index=False)

        def _mock_download_minute_file(code, f, *args, **kwargs):
            prices = pd.concat(
                dict(
                    Close=pd.DataFrame(
                        dict(
                            FI12345=[
                                48.90,
                                49.40,
                                55.49,
                                56.78
                            ],
                            FI23456=[
                                59.5,
                                59.59,
                                59.34,
                                51.56,
                            ]
                        ),
                        index=[
                            "2018-04-01T09:30:00-04:00",
                            "2018-04-01T15:30:00-04:00",
                            "2018-04-02T09:30:00-04:00",
                            "2018-04-02T15:30:00-04:00",
                            ])
                    ), names=["Field","Date"]
            )
            prices.to_csv(f)

        def mock_download_master_file(f, *args, **kwargs):
            securities = pd.DataFrame(dict(Sid=["FI12345","FI23456"],
                                           Timezone=["America/New_York", "America/New_York"]))
            securities.to_csv(f, index=False)
            f.seek(0)

        def mock_list_history_databases():
            return [
                "usa-stk-1min",
                "japan-stk-1d",
                "demo-stk-1min"
            ]

        def mock_list_realtime_databases():
            return {"usa-stk-snapshot": ["usa-stk-snapshot-1min"]}

        def mock_list_bundles():
            return {"usstock-1min": True, "usstock-free-1min": True}

        mock_download_history_file.side_effect = _mock_download_history_file
        mock_download_market_data_file.side_effect = _mock_download_market_data_file
        mock_download_minute_file.side_effect = _mock_download_minute_file

        with patch('quantrocket.price.list_bundles', new=mock_list_bundles):
            with patch('quantrocket.price.list_realtime_databases', new=mock_list_realtime_databases):
                with patch('quantrocket.price.list_history_databases', new=mock_list_history_databases):
                    with patch('quantrocket.price.get_history_db_config', new=mock_get_history_db_config):
                        with patch('quantrocket.price.get_realtime_db_config', new=mock_get_realtime_db_config):
                            with patch("quantrocket.price.download_master_file", new=mock_download_master_file):

                                prices = get_prices(
                                    ["usa-stk-1min", "usa-stk-snapshot-1min", "usstock-1min"],
                                    fields=["Close", "LastClose", "Volume", "Wap"],
                                    times=["09:30:00", "15:30:00"],
                                )

        self.assertEqual(len(mock_download_history_file.mock_calls), 1)
        history_call = mock_download_history_file.mock_calls[0]
        _, args, kwargs = history_call
        self.assertEqual(args[0], "usa-stk-1min")
        # only supported subset of fields is requested
        self.assertSetEqual(set(kwargs["fields"]), {"Wap", "Volume"})
        # times filter was passed
        self.assertListEqual(kwargs["times"], ["09:30:00", "15:30:00"])

        self.assertEqual(len(mock_download_market_data_file.mock_calls), 1)
        realtime_call = mock_download_market_data_file.mock_calls[0]
        _, args, kwargs = realtime_call
        self.assertEqual(args[0], "usa-stk-snapshot-1min")
        # only supported subset of fields is requested
        self.assertListEqual(kwargs["fields"], ["LastClose"])
        # times filter not passed
        self.assertNotIn("times", list(kwargs.keys()))

        self.assertEqual(len(mock_download_minute_file.mock_calls), 1)
        minute_call = mock_download_minute_file.mock_calls[0]
        _, args, kwargs = minute_call
        self.assertEqual(args[0], "usstock-1min")
        # only supported subset of fields is requested
        self.assertSetEqual(set(kwargs["fields"]), {"Close", "Volume"})
        # times filter was passed
        self.assertListEqual(kwargs["times"], ["09:30:00", "15:30:00"])

        self.assertListEqual(list(prices.index.names), ["Field", "Date", "Time"])
        self.assertEqual(prices.columns.name, "Sid")
        dt = prices.index.get_level_values("Date")
        self.assertTrue(isinstance(dt, pd.DatetimeIndex))
        self.assertIsNone(dt.tz)
        self.assertListEqual(list(prices.columns), ["FI12345","FI23456"])
        waps = prices.loc["Wap"]
        waps = waps.reset_index()
        waps.loc[:, "Date"] = waps.Date.dt.strftime("%Y-%m-%dT%H:%M:%S%z")
        self.assertListEqual(
            waps.fillna("nan").to_dict(orient="records"),
            [{'Date': '2018-04-01T00:00:00', 'Time': '09:30:00', "FI12345": 20.1, "FI23456": 50.5},
             {'Date': '2018-04-01T00:00:00', 'Time': '15:30:00', "FI12345": 20.5, "FI23456": 52.5},
             {'Date': '2018-04-02T00:00:00', 'Time': '09:30:00', "FI12345": 19.4, "FI23456": 51.59},
             {'Date': '2018-04-02T00:00:00', 'Time': '15:30:00', "FI12345": 18.56, "FI23456": 54.23}]
        )
        volumes = prices.loc["Volume"]
        volumes = volumes.reset_index()
        volumes.loc[:, "Date"] = volumes.Date.dt.strftime("%Y-%m-%dT%H:%M:%S%z")
        self.assertListEqual(
            volumes.to_dict(orient="records"),
            [{'Date': '2018-04-01T00:00:00', 'Time': '09:30:00', "FI12345": 15000.0, "FI23456": 98000.0},
             {'Date': '2018-04-01T00:00:00', 'Time': '15:30:00', "FI12345": 7800.0, "FI23456": 179000.0},
             {'Date': '2018-04-02T00:00:00', 'Time': '09:30:00', "FI12345": 12400.0, "FI23456": 142500.0},
             {'Date': '2018-04-02T00:00:00', 'Time': '15:30:00', "FI12345": 14500.0, "FI23456": 124000.0}]
        )

        last_closes = prices.loc["LastClose"]
        last_closes = last_closes.reset_index()
        last_closes.loc[:, "Date"] = last_closes.Date.dt.strftime("%Y-%m-%dT%H:%M:%S%z")
        self.assertListEqual(
            last_closes.fillna("nan").to_dict(orient="records"),
            # Data was UTC but now NY
            [{'Date': '2018-04-01T00:00:00', 'Time': '09:30:00', "FI12345": 'nan', "FI23456": 'nan'},
             {'Date': '2018-04-01T00:00:00', 'Time': '15:30:00', "FI12345": 39.4, "FI23456": 79.59},
             {'Date': '2018-04-02T00:00:00', 'Time': '09:30:00', "FI12345": 'nan', "FI23456": 'nan'},
             {'Date': '2018-04-02T00:00:00', 'Time': '15:30:00', "FI12345": 46.78, "FI23456": 81.56}]
        )

        closes = prices.loc["Close"]
        closes = closes.reset_index()
        closes.loc[:, "Date"] = closes.Date.dt.strftime("%Y-%m-%dT%H:%M:%S%z")
        self.assertListEqual(
            closes.fillna("nan").to_dict(orient="records"),
            [{'Date': '2018-04-01T00:00:00', 'Time': '09:30:00', 'FI12345': 48.9, 'FI23456': 59.5},
             {'Date': '2018-04-01T00:00:00', 'Time': '15:30:00', 'FI12345': 49.4, 'FI23456': 59.59},
             {'Date': '2018-04-02T00:00:00', 'Time': '09:30:00', 'FI12345': 55.49, 'FI23456': 59.34},
             {'Date': '2018-04-02T00:00:00', 'Time': '15:30:00', 'FI12345': 56.78, 'FI23456': 51.56}]
        )

    def test_consolidate_overlapping_fields_and_respect_priority(self):
        """
        Tests that when querying two databases with overlapping
        dates/sids/fields, the value is taken from the db which was passed
        first as an argument.
        """
        def mock_get_history_db_config(db):
            if db == "usa-stk-1d":
                return {
                    "bar_size": "1 day",
                    "universes": ["usa-stk"],
                    "vendor": "ibkr",
                    "fields": ["Close","Open","High","Low", "Volume"]
                }
            else:
                return {
                    "bar_size": "1 day",
                    "universes": ["nyse-stk"],
                    "vendor": "ibkr",
                    "fields": ["Close","Open","High","Low", "Volume"]
                }

        def mock_download_history_file(code, f, *args, **kwargs):
            if code == "usa-stk-1d":
                prices = pd.DataFrame(
                    dict(
                        Sid=[
                            "FI12345",
                            "FI12345",
                            "FI12345",
                            "FI23456",
                            "FI23456",
                            "FI23456",
                            ],
                        Date=[
                            "2018-04-01",
                            "2018-04-02",
                            "2018-04-03",
                            "2018-04-01",
                            "2018-04-02",
                            "2018-04-03"
                            ],
                        Close=[
                            20.10,
                            20.50,
                            19.40,
                            50.5,
                            52.5,
                            51.59,
                            ],
                        Volume=[
                            15000,
                            7800,
                            12400,
                            98000,
                            179000,
                            142500
                        ]
                    ))
            else:
                prices = pd.DataFrame(
                    dict(
                        Sid=[
                            "FI12345",
                            "FI12345",
                            "FI12345",
                            ],
                        Date=[
                            "2018-04-01",
                            "2018-04-02",
                            "2018-04-03",
                            ],
                        Close=[
                            5900,
                            5920,
                            5950],
                    ))
            prices.to_csv(f, index=False)

        def mock_list_history_databases():
            return [
                "usa-stk-1d",
                "nyse-stk-1d",
                "usa-stk-15min",
                "demo-stk-1min"
            ]

        def mock_list_realtime_databases():
            return {"demo-stk-taq": ["demo-stk-taq-1h"]}

        def mock_list_bundles():
            return {"usstock-1min": True}

        with patch('quantrocket.price.list_bundles', new=mock_list_bundles):
            with patch('quantrocket.price.list_realtime_databases', new=mock_list_realtime_databases):
                with patch('quantrocket.price.list_history_databases', new=mock_list_history_databases):
                    with patch('quantrocket.price.get_history_db_config', new=mock_get_history_db_config):
                        with patch('quantrocket.price.download_history_file', new=mock_download_history_file):

                            # prioritize usa-stk-1d by passing first
                            prices = get_prices(["usa-stk-1d", "nyse-stk-1d"],
                                                fields=["Close", "Volume"])

        self.assertListEqual(list(prices.index.names), ["Field", "Date"])
        self.assertEqual(prices.columns.name, "Sid")
        dt = prices.index.get_level_values("Date")
        self.assertTrue(isinstance(dt, pd.DatetimeIndex))
        self.assertIsNone(dt.tz)
        self.assertListEqual(list(prices.columns), ["FI12345","FI23456"])
        closes = prices.loc["Close"]
        closes = closes.reset_index()
        closes.loc[:, "Date"] = closes.Date.dt.strftime("%Y-%m-%dT%H:%M:%S%z")
        self.assertListEqual(
            closes.to_dict(orient="records"),
            [{'Date': '2018-04-01T00:00:00', "FI12345": 20.1, "FI23456": 50.5},
             {'Date': '2018-04-02T00:00:00', "FI12345": 20.5, "FI23456": 52.5},
             {'Date': '2018-04-03T00:00:00', "FI12345": 19.4, "FI23456": 51.59}]
        )
        volumes = prices.loc["Volume"]
        volumes = volumes.reset_index()
        volumes.loc[:, "Date"] = volumes.Date.dt.strftime("%Y-%m-%dT%H:%M:%S%z")
        self.assertListEqual(
            volumes.to_dict(orient="records"),
            [{'Date': '2018-04-01T00:00:00', "FI12345": 15000, "FI23456": 98000},
             {'Date': '2018-04-02T00:00:00', "FI12345": 7800, "FI23456": 179000},
             {'Date': '2018-04-03T00:00:00', "FI12345": 12400, "FI23456": 142500}]
        )

        # repeat test but prioritize nyse-stk-1d by passing first
        with patch('quantrocket.price.list_bundles', new=mock_list_bundles):
            with patch('quantrocket.price.list_realtime_databases', new=mock_list_realtime_databases):
                with patch('quantrocket.price.list_history_databases', new=mock_list_history_databases):
                    with patch('quantrocket.price.get_history_db_config', new=mock_get_history_db_config):
                        with patch('quantrocket.price.download_history_file', new=mock_download_history_file):

                            prices = get_prices(["nyse-stk-1d", "usa-stk-1d"],
                                                fields=["Close", "Volume"])

        self.assertListEqual(list(prices.index.names), ["Field", "Date"])
        self.assertEqual(prices.columns.name, "Sid")
        dt = prices.index.get_level_values("Date")
        self.assertTrue(isinstance(dt, pd.DatetimeIndex))
        self.assertIsNone(dt.tz)
        self.assertListEqual(list(prices.columns), ["FI12345","FI23456"])
        closes = prices.loc["Close"]
        closes = closes.reset_index()
        closes.loc[:, "Date"] = closes.Date.dt.strftime("%Y-%m-%dT%H:%M:%S%z")
        self.assertListEqual(
            closes.to_dict(orient="records"),
            [{'Date': '2018-04-01T00:00:00', "FI12345": 5900.0, "FI23456": 50.5},
             {'Date': '2018-04-02T00:00:00', "FI12345": 5920.0, "FI23456": 52.5},
             {'Date': '2018-04-03T00:00:00', "FI12345": 5950.0, "FI23456": 51.59}]
        )
        volumes = prices.loc["Volume"]
        volumes = volumes.reset_index()
        volumes.loc[:, "Date"] = volumes.Date.dt.strftime("%Y-%m-%dT%H:%M:%S%z")
        # Since volume is null in nyse-stk-1d, we get the volume from usa-stk-1d
        self.assertListEqual(
            volumes.to_dict(orient="records"),
            [{'Date': '2018-04-01T00:00:00', "FI12345": 15000, "FI23456": 98000},
             {'Date': '2018-04-02T00:00:00', "FI12345": 7800, "FI23456": 179000},
             {'Date': '2018-04-03T00:00:00', "FI12345": 12400, "FI23456": 142500}]
        )

    def test_parse_bar_sizes(self):
        """
        Tests that when querying a history and real-time database which have
        different bar size strings which are nevertheless equivalent, this is
        allowed.
        """
        def mock_get_history_db_config(db):
            return {
                "bar_size": "15 mins", # 15 mins plural (IB format)
                "universes": ["usa-stk"],
                "vendor": "ibkr",
                "fields": ["Close","Open","High","Low", "Volume"]
            }

        def mock_get_realtime_db_config(db):
            return {
                "bar_size": "15 min", # 15 min Pandas format
                "fields": ["LastClose","LastOpen","LastHigh","LastLow", "VolumeClose"]
            }
        def mock_download_history_file(code, f, *args, **kwargs):
            prices = pd.DataFrame(
                dict(
                    Sid=[
                        "FI12345",
                        "FI12345",
                        "FI12345",
                        "FI12345",
                        "FI23456",
                        "FI23456",
                        "FI23456",
                        "FI23456",
                        ],
                    Date=[
                        "2018-04-01T09:30:00-04:00",
                        "2018-04-01T15:30:00-04:00",
                        "2018-04-02T09:30:00-04:00",
                        "2018-04-02T15:30:00-04:00",
                        "2018-04-01T09:30:00-04:00",
                        "2018-04-01T15:30:00-04:00",
                        "2018-04-02T09:30:00-04:00",
                        "2018-04-02T15:30:00-04:00",
                        ],
                    Close=[
                        20.10,
                        20.50,
                        19.40,
                        18.56,
                        50.5,
                        52.5,
                        51.59,
                        54.23
                        ],
                    Volume=[
                        15000,
                        7800,
                        12400,
                        14500,
                        98000,
                        179000,
                        142500,
                        124000,
                    ]
                ))
            prices.to_csv(f, index=False)

        def mock_download_market_data_file(code, f, *args, **kwargs):
            prices = pd.DataFrame(
                dict(
                    Sid=[
                        "FI12345",
                        "FI12345",
                        "FI23456",
                        "FI23456",
                        ],
                    Date=[
                        "2018-04-01T19:30:00+00",
                        "2018-04-02T19:30:00+00",
                        "2018-04-01T19:30:00+00",
                        "2018-04-02T19:30:00+00",
                        ],
                    LastClose=[
                        30.50,
                        39.40,
                        79.5,
                        79.59,
                        ]
                ))
            prices.to_csv(f, index=False)

        def mock_download_master_file(f, *args, **kwargs):
            securities = pd.DataFrame(dict(Sid=["FI12345","FI23456"],
                                           Timezone=["America/New_York", "America/New_York"]))
            securities.to_csv(f, index=False)
            f.seek(0)

        def mock_list_history_databases():
            return [
                "usa-stk-15min",
                "japan-stk-1d",
                "demo-stk-1min"
            ]

        def mock_list_realtime_databases():
            return {"usa-stk-snapshot": ["usa-stk-snapshot-15min"]}

        def mock_list_bundles():
            return {"usstock-1min": True}

        with patch('quantrocket.price.list_bundles', new=mock_list_bundles):
            with patch('quantrocket.price.list_realtime_databases', new=mock_list_realtime_databases):
                with patch('quantrocket.price.list_history_databases', new=mock_list_history_databases):
                    with patch('quantrocket.price.get_history_db_config', new=mock_get_history_db_config):
                        with patch('quantrocket.price.get_realtime_db_config', new=mock_get_realtime_db_config):
                            with patch('quantrocket.price.download_history_file', new=mock_download_history_file):
                                with patch('quantrocket.price.download_market_data_file', new=mock_download_market_data_file):
                                    with patch("quantrocket.price.download_master_file", new=mock_download_master_file):

                                        prices = get_prices(
                                            ["usa-stk-15min", "usa-stk-snapshot-15min"],
                                            fields=["Close", "LastClose", "Volume"])

        self.assertListEqual(list(prices.index.names), ["Field", "Date", "Time"])
        self.assertEqual(prices.columns.name, "Sid")
        dt = prices.index.get_level_values("Date")
        self.assertTrue(isinstance(dt, pd.DatetimeIndex))
        self.assertIsNone(dt.tz)
        self.assertListEqual(list(prices.columns), ["FI12345","FI23456"])
        closes = prices.loc["Close"]
        closes = closes.reset_index()
        closes.loc[:, "Date"] = closes.Date.dt.strftime("%Y-%m-%dT%H:%M:%S%z")
        self.assertListEqual(
            closes.to_dict(orient="records"),
            [{'Date': '2018-04-01T00:00:00', 'Time': '09:30:00', "FI12345": 20.1, "FI23456": 50.5},
             {'Date': '2018-04-01T00:00:00', 'Time': '15:30:00', "FI12345": 20.5, "FI23456": 52.5},
             {'Date': '2018-04-02T00:00:00', 'Time': '09:30:00', "FI12345": 19.4, "FI23456": 51.59},
             {'Date': '2018-04-02T00:00:00', 'Time': '15:30:00', "FI12345": 18.56, "FI23456": 54.23}]
        )
        volumes = prices.loc["Volume"]
        volumes = volumes.reset_index()
        volumes.loc[:, "Date"] = volumes.Date.dt.strftime("%Y-%m-%dT%H:%M:%S%z")
        self.assertListEqual(
            volumes.to_dict(orient="records"),
            [{'Date': '2018-04-01T00:00:00', 'Time': '09:30:00', "FI12345": 15000.0, "FI23456": 98000.0},
             {'Date': '2018-04-01T00:00:00', 'Time': '15:30:00', "FI12345": 7800.0, "FI23456": 179000.0},
             {'Date': '2018-04-02T00:00:00', 'Time': '09:30:00', "FI12345": 12400.0, "FI23456": 142500.0},
             {'Date': '2018-04-02T00:00:00', 'Time': '15:30:00', "FI12345": 14500.0, "FI23456": 124000.0}]
        )

        last_closes = prices.loc["LastClose"]
        last_closes = last_closes.reset_index()
        last_closes.loc[:, "Date"] = last_closes.Date.dt.strftime("%Y-%m-%dT%H:%M:%S%z")
        self.assertListEqual(
            last_closes.fillna("nan").to_dict(orient="records"),
            [{'Date': '2018-04-01T00:00:00', 'Time': '09:30:00', "FI12345": "nan", "FI23456": "nan"},
             {'Date': '2018-04-01T00:00:00', 'Time': '15:30:00', "FI12345": 30.5, "FI23456": 79.5},
             {'Date': '2018-04-02T00:00:00', 'Time': '09:30:00', "FI12345": "nan", "FI23456": "nan"},
             {'Date': '2018-04-02T00:00:00', 'Time': '15:30:00', "FI12345": 39.4, "FI23456": 79.59}]
        )

    def test_no_complain_no_data_multiple_dbs(self):
        """
        Tests that if multiple dbs are queried and one lacks data but the
        other has data, no error is raised.
        """
        def mock_get_history_db_config(db):
            if db == "usa-stk-1d":
                return {
                    "bar_size": "1 day",
                    "universes": ["usa-stk"],
                    "vendor": "ibkr",
                "fields": ["Close","Open","High","Low", "Volume"]
                }
            else:
                return {
                    "bar_size": "1 day",
                    "universes": ["japan-stk"],
                    "vendor": "ibkr",
                "fields": ["Close","Open","High","Low", "Volume"]
                }

        def mock_download_history_file(code, f, *args, **kwargs):
            if code == "usa-stk-1d":
                prices = pd.DataFrame(
                    dict(
                        Sid=[
                            "FI12345",
                            "FI12345",
                            "FI12345",
                            "FI23456",
                            "FI23456",
                            "FI23456",
                            ],
                        Date=[
                            "2018-04-01",
                            "2018-04-02",
                            "2018-04-03",
                            "2018-04-01",
                            "2018-04-02",
                            "2018-04-03"
                            ],
                        Close=[
                            20.10,
                            20.50,
                            19.40,
                            50.5,
                            52.5,
                            51.59,
                            ],
                        Volume=[
                            15000,
                            7800,
                            12400,
                            98000,
                            179000,
                            142500
                        ]
                    ))
                prices.to_csv(f, index=False)
            else:
                raise NoHistoricalData("no history matches the query parameters")

        def mock_list_history_databases():
            return [
                "usa-stk-1d",
                "japan-stk-1d",
                "usa-stk-15min",
                "demo-stk-1min"
            ]

        def mock_list_realtime_databases():
            return {"demo-stk-taq": ["demo-stk-taq-1h"]}

        def mock_list_bundles():
            return {"usstock-1min": True}

        with patch('quantrocket.price.list_bundles', new=mock_list_bundles):
            with patch('quantrocket.price.list_realtime_databases', new=mock_list_realtime_databases):
                with patch('quantrocket.price.list_history_databases', new=mock_list_history_databases):
                    with patch('quantrocket.price.get_history_db_config', new=mock_get_history_db_config):
                        with patch('quantrocket.price.download_history_file', new=mock_download_history_file):

                            prices = get_prices(["usa-stk-1d", "japan-stk-1d"], fields=["Close", "Volume"])

        self.assertListEqual(list(prices.index.names), ["Field", "Date"])
        self.assertEqual(prices.columns.name, "Sid")
        dt = prices.index.get_level_values("Date")
        self.assertTrue(isinstance(dt, pd.DatetimeIndex))
        self.assertIsNone(dt.tz) # EOD is tz-naive
        self.assertListEqual(list(prices.columns), ["FI12345","FI23456"])
        closes = prices.loc["Close"]
        closes = closes.reset_index()
        closes.loc[:, "Date"] = closes.Date.dt.strftime("%Y-%m-%dT%H:%M:%S%z")
        self.assertListEqual(
            closes.to_dict(orient="records"),
            [{'Date': '2018-04-01T00:00:00', "FI12345": 20.1, "FI23456": 50.5},
             {'Date': '2018-04-02T00:00:00', "FI12345": 20.5, "FI23456": 52.5},
             {'Date': '2018-04-03T00:00:00', "FI12345": 19.4, "FI23456": 51.59}]
        )
        volumes = prices.loc["Volume"]
        volumes = volumes.reset_index()
        volumes.loc[:, "Date"] = volumes.Date.dt.strftime("%Y-%m-%dT%H:%M:%S%z")
        self.assertListEqual(
            volumes.to_dict(orient="records"),
            [{'Date': '2018-04-01T00:00:00', "FI12345": 15000, "FI23456": 98000},
             {'Date': '2018-04-02T00:00:00', "FI12345": 7800, "FI23456": 179000},
             {'Date': '2018-04-03T00:00:00', "FI12345": 12400, "FI23456": 142500}]
        )

    def test_complain_no_data_multiple_dbs(self):
        """
        Tests that if multiple dbs are queried and all of them lack data, an
        error is raised.
        """
        def mock_get_history_db_config(db):
            if db == "usa-stk-1d":
                return {
                    "bar_size": "1 day",
                    "universes": ["usa-stk"],
                    "vendor": "ibkr",
                    "fields": ["Close","Open","High","Low", "Volume"]
                }
            else:
                return {
                    "bar_size": "1 day",
                    "universes": ["japan-stk"],
                    "vendor": "ibkr",
                    "fields": ["Close","Open","High","Low", "Volume"]
                }

        def mock_download_history_file(code, f, *args, **kwargs):
            raise NoHistoricalData("no history matches the query parameters")

        def mock_list_history_databases():
            return [
                "usa-stk-1d",
                "japan-stk-1d",
                "usa-stk-15min",
                "demo-stk-1min"
            ]

        def mock_list_realtime_databases():
            return {"demo-stk-taq": ["demo-stk-taq-1h"]}

        def mock_list_bundles():
            return {"usstock-1min": True}

        with patch('quantrocket.price.list_bundles', new=mock_list_bundles):
            with patch('quantrocket.price.list_realtime_databases', new=mock_list_realtime_databases):
                with patch('quantrocket.price.list_history_databases', new=mock_list_history_databases):
                    with patch('quantrocket.price.get_history_db_config', new=mock_get_history_db_config):
                        with patch('quantrocket.price.download_history_file', new=mock_download_history_file):

                            with self.assertRaises(NoHistoricalData) as cm:
                                get_prices(["usa-stk-1d", "japan-stk-1d"])

        self.assertIn(
            "no price data matches the query parameters in any of usa-stk-1d, japan-stk-1d",
            str(cm.exception))

    def test_complain_no_data_single_db(self):
        """
        Tests that if a single db is queried and it lacks data, an
        error is raised.
        """
        def mock_get_history_db_config(db):
            return {
                "bar_size": "1 day",
                "universes": ["usa-stk"],
                "vendor": "ibkr",
                "fields": ["Close","Open","High","Low", "Volume"]
            }

        def mock_download_history_file(code, f, *args, **kwargs):
            raise NoHistoricalData("this error will be passed through as is")


        def mock_list_history_databases():
            return [
                "usa-stk-1d",
                "japan-stk-1d",
                "usa-stk-15min",
                "demo-stk-1min"
            ]

        def mock_list_realtime_databases():
            return {"demo-stk-taq": ["demo-stk-taq-1h"]}

        def mock_list_bundles():
            return {"usstock-1min": True}

        with patch('quantrocket.price.list_bundles', new=mock_list_bundles):
            with patch('quantrocket.price.list_realtime_databases', new=mock_list_realtime_databases):
                with patch('quantrocket.price.list_history_databases', new=mock_list_history_databases):
                    with patch('quantrocket.price.get_history_db_config', new=mock_get_history_db_config):
                        with patch('quantrocket.price.download_history_file', new=mock_download_history_file):

                            with self.assertRaises(NoHistoricalData) as cm:
                                get_prices("usa-stk-1d")

        self.assertIn(
            "this error will be passed through as is",
            str(cm.exception))

    def test_complain_if_tick_db(self):
        """
        Tests error handling when multiple dbs are queried and one is a tick
        db rather than a history or real-time agg db.
        """
        def mock_list_history_databases():
            return [
                "usa-stk-1d",
                "japan-stk-1d",
                "usa-stk-15min",
                "demo-stk-1min",
                "japan-stk-15min"
            ]

        def mock_list_realtime_databases():
            return {"demo-stk-taq": ["demo-stk-taq-1h"],
                    "etf-taq": ["etf-taq-1h"],
                    }

        def mock_list_bundles():
            return {"usstock-1min": True}

        with patch('quantrocket.price.list_bundles', new=mock_list_bundles):
            with patch('quantrocket.price.list_realtime_databases', new=mock_list_realtime_databases):
                with patch('quantrocket.price.list_history_databases', new=mock_list_history_databases):

                        with self.assertRaises(ParameterError) as cm:
                            get_prices(["usa-stk-1d", "etf-taq", "japan-stk-15min"])

        self.assertIn((
            "etf-taq is a real-time tick database, only history databases or "
            "real-time aggregate databases are supported"
            ), str(cm.exception))

    def test_complain_if_unknown_db(self):
        """
        Tests error handling when an unknown db is queried.
        """
        def mock_list_history_databases():
            return [
                "usa-stk-1d",
                "japan-stk-1d",
                "usa-stk-15min",
                "demo-stk-1min",
                "japan-stk-15min"
            ]

        def mock_list_realtime_databases():
            return {"demo-stk-taq": ["demo-stk-taq-1h"],
                    "etf-taq": ["etf-taq-1h"],
                    }

        def mock_list_bundles():
            return {"usstock-1min": True}

        with patch('quantrocket.price.list_bundles', new=mock_list_bundles):
            with patch('quantrocket.price.list_realtime_databases', new=mock_list_realtime_databases):
                with patch('quantrocket.price.list_history_databases', new=mock_list_history_databases):

                        with self.assertRaises(ParameterError) as cm:
                            get_prices(["asx-stk-1d"])

        self.assertIn((
           "no history or real-time aggregate databases or Zipline bundles called asx-stk-1d"
            ), str(cm.exception))

    def test_complain_if_multiple_bar_sizes(self):
        """
        Tests error handling when multiple dbs are queried and they have
        different bar sizes.
        """
        def mock_get_history_db_config(db):
            if db == "usa-stk-1d":
                return {
                    "bar_size": "1 day",
                    "universes": ["usa-stk"],
                    "vendor": "ibkr",
                    "fields": ["Close","Open","High","Low", "Volume"]
                }
            else:
                return {
                    "bar_size": "30 mins",
                    "universes": ["japan-stk"],
                    "vendor": "ibkr",
                    "fields": ["Close","Open","High","Low", "Volume"]
                }

        def mock_list_history_databases():
            return [
                "usa-stk-1d",
                "japan-stk-1d",
                "usa-stk-15min",
                "demo-stk-1min",
                "japan-stk-15min"
            ]

        def mock_list_realtime_databases():
            return {"demo-stk-taq": ["demo-stk-taq-1h"]}

        def mock_list_bundles():
            return {"usstock-1min": True}

        with patch('quantrocket.price.list_realtime_databases', new=mock_list_realtime_databases):
            with patch('quantrocket.price.list_history_databases', new=mock_list_history_databases):
                with patch('quantrocket.price.get_history_db_config', new=mock_get_history_db_config):
                    with patch('quantrocket.price.list_bundles', new=mock_list_bundles):
                        with self.assertRaises(ParameterError) as cm:
                            get_prices(["usa-stk-1d", "japan-stk-15min", "usstock-1min"])

        self.assertIn((
            "all databases must contain same bar size but usa-stk-1d, japan-stk-15min, usstock-1min have different "
            "bar sizes:"
            ), str(cm.exception))

    def test_intraday_pass_timezone(self):
        """
        Tests handling of an intraday db when a timezone is specified.
        """
        def mock_get_history_db_config(db):
            return {
                "bar_size": "30 mins",
                "universes": ["usa-stk"],
                "vendor": "ibkr",
                "fields": ["Close","Open","High","Low", "Volume"]
            }

        def mock_download_history_file(code, f, *args, **kwargs):
            prices = pd.DataFrame(
                dict(
                    Sid=[
                        "FI12345",
                        "FI12345",
                        "FI12345",
                        "FI12345",
                        "FI23456",
                        "FI23456",
                        "FI23456",
                        "FI23456"
                        ],
                    Date=[
                        "2018-04-01T09:30:00-04:00",
                        "2018-04-01T10:00:00-04:00",
                        "2018-04-02T09:30:00-04:00",
                        "2018-04-02T10:00:00-04:00",
                        "2018-04-01T09:30:00-04:00",
                        "2018-04-01T10:00:00-04:00",
                        "2018-04-02T09:30:00-04:00",
                        "2018-04-02T10:00:00-04:00",
                        ],
                    Close=[
                        20.10,
                        20.25,
                        20.50,
                        20.38,
                        50.15,
                        50.59,
                        51.59,
                        51.67
                        ],
                    Volume=[
                        1500,
                        7800,
                        1400,
                        800,
                        9000,
                        7100,
                        1400,
                        1500
                    ]
                ))
            prices.to_csv(f, index=False)
        def mock_list_history_databases():
            return [
                "usa-stk-1d",
                "japan-stk-1d",
                "usa-stk-15min",
                "demo-stk-1min"
            ]

        def mock_list_realtime_databases():
            return {"demo-stk-taq": ["demo-stk-taq-1h"]}

        def mock_list_bundles():
            return {"usstock-1min": True}

        with patch('quantrocket.price.list_bundles', new=mock_list_bundles):
            with patch('quantrocket.price.list_realtime_databases', new=mock_list_realtime_databases):
                with patch('quantrocket.price.list_history_databases', new=mock_list_history_databases):
                    with patch('quantrocket.price.get_history_db_config', new=mock_get_history_db_config):
                        with patch('quantrocket.price.download_history_file', new=mock_download_history_file):

                            prices = get_prices(
                                "usa-stk-15min",
                                times=["09:30:00", "10:00:00"],
                                fields=["Close", "Volume"],
                                timezone="America/Chicago"
                            )

        self.assertListEqual(list(prices.index.names), ["Field", "Date", "Time"])
        self.assertEqual(prices.columns.name, "Sid")
        dt = prices.index.get_level_values("Date")
        self.assertTrue(isinstance(dt, pd.DatetimeIndex))
        self.assertIsNone(dt.tz)
        self.assertListEqual(list(prices.columns), ["FI12345","FI23456"])
        self.assertSetEqual(
            set(prices.index.get_level_values("Field")),
            {"Close", "Volume"})
        self.assertSetEqual(
            set(prices.index.get_level_values("Time")),
            {"08:30:00", "09:00:00"})
        closes = prices.loc["Close"]
        closes_830 = closes.xs("08:30:00", level="Time")
        closes_830 = closes_830.reset_index()
        closes_830.loc[:, "Date"] = closes_830.Date.dt.strftime("%Y-%m-%dT%H:%M:%S%z")
        self.assertListEqual(
            closes_830.to_dict(orient="records"),
            [{'Date': '2018-04-01T00:00:00', "FI12345": 20.1, "FI23456": 50.15},
             {'Date': '2018-04-02T00:00:00', "FI12345": 20.5, "FI23456": 51.59}]
        )
        closes_900 = closes.xs("09:00:00", level="Time")
        closes_900 = closes_900.reset_index()
        closes_900.loc[:, "Date"] = closes_900.Date.dt.strftime("%Y-%m-%dT%H:%M:%S%z")
        self.assertListEqual(
            closes_900.to_dict(orient="records"),
            [{'Date': '2018-04-01T00:00:00', "FI12345": 20.25, "FI23456": 50.59},
             {'Date': '2018-04-02T00:00:00', "FI12345": 20.38, "FI23456": 51.67}]
        )

        volumes = prices.loc["Volume"]
        volumes_830 = volumes.xs("08:30:00", level="Time")
        volumes_830 = volumes_830.reset_index()
        volumes_830.loc[:, "Date"] = volumes_830.Date.dt.strftime("%Y-%m-%dT%H:%M:%S%z")
        self.assertListEqual(
            volumes_830.to_dict(orient="records"),
            [{'Date': '2018-04-01T00:00:00', "FI12345": 1500.0, "FI23456": 9000.0},
             {'Date': '2018-04-02T00:00:00', "FI12345": 1400.0, "FI23456": 1400.0}]
        )

    def test_intraday_infer_timezone_from_securities(self):
        """
        Tests handling of an intraday db when a timezone is not specified and
        is therefore inferred from the securities master.
        """
        def mock_get_history_db_config(db):
            return {
                "bar_size": "30 mins",
                "universes": ["usa-stk"],
                "vendor": "ibkr"
            }

        def mock_download_history_file(code, f, *args, **kwargs):
            prices = pd.DataFrame(
                dict(
                    Sid=[
                        "FI12345",
                        "FI12345",
                        "FI12345",
                        "FI12345",
                        "FI23456",
                        "FI23456",
                        "FI23456",
                        "FI23456"
                        ],
                    Date=[
                        "2018-04-01T09:30:00-04:00",
                        "2018-04-01T10:00:00-04:00",
                        "2018-04-02T09:30:00-04:00",
                        "2018-04-02T10:00:00-04:00",
                        "2018-04-01T09:30:00-04:00",
                        "2018-04-01T10:00:00-04:00",
                        "2018-04-02T09:30:00-04:00",
                        "2018-04-02T10:00:00-04:00",
                        ],
                    Close=[
                        20.10,
                        20.25,
                        20.50,
                        20.38,
                        50.15,
                        50.59,
                        51.59,
                        51.67
                        ],
                    Volume=[
                        1500,
                        7800,
                        1400,
                        800,
                        9000,
                        7100,
                        1400,
                        1500
                    ]
                ))
            prices.to_csv(f, index=False)

        def mock_download_master_file(f, *args, **kwargs):
            securities = pd.DataFrame(dict(Sid=["FI12345","FI23456"],
                                           Timezone=["America/New_York", "America/New_York"]))
            securities.to_csv(f, index=False)
            f.seek(0)

        def mock_list_history_databases():
            return [
                "usa-stk-1d",
                "japan-stk-1d",
                "usa-stk-15min",
                "demo-stk-1min"
            ]

        def mock_list_realtime_databases():
            return {"demo-stk-taq": ["demo-stk-taq-1h"]}

        def mock_list_bundles():
            return {"usstock-1min": True}

        with patch('quantrocket.price.list_bundles', new=mock_list_bundles):
            with patch('quantrocket.price.list_realtime_databases', new=mock_list_realtime_databases):
                with patch('quantrocket.price.list_history_databases', new=mock_list_history_databases):
                    with patch('quantrocket.price.get_history_db_config', new=mock_get_history_db_config):
                        with patch('quantrocket.price.download_history_file', new=mock_download_history_file):
                            with patch("quantrocket.price.download_master_file", new=mock_download_master_file):

                                prices = get_prices(
                                    "usa-stk-15min",
                                    times=["09:30:00", "10:00:00"],
                                    fields=["Close", "Volume"]
                                )

        self.assertListEqual(list(prices.index.names), ["Field", "Date", "Time"])
        self.assertEqual(prices.columns.name, "Sid")
        dt = prices.index.get_level_values("Date")
        self.assertTrue(isinstance(dt, pd.DatetimeIndex))
        self.assertIsNone(dt.tz)
        self.assertListEqual(list(prices.columns), ["FI12345","FI23456"])
        self.assertSetEqual(
            set(prices.index.get_level_values("Field")),
            {"Close", "Volume"})
        self.assertSetEqual(
            set(prices.index.get_level_values("Time")),
            {"09:30:00", "10:00:00"})
        closes = prices.loc["Close"]
        closes_930 = closes.xs("09:30:00", level="Time")
        closes_930 = closes_930.reset_index()
        closes_930.loc[:, "Date"] = closes_930.Date.dt.strftime("%Y-%m-%dT%H:%M:%S%z")
        self.assertListEqual(
            closes_930.to_dict(orient="records"),
            [{'Date': '2018-04-01T00:00:00', "FI12345": 20.1, "FI23456": 50.15},
             {'Date': '2018-04-02T00:00:00', "FI12345": 20.5, "FI23456": 51.59}]
        )
        closes_1000 = closes.xs("10:00:00", level="Time")
        closes_1000 = closes_1000.reset_index()
        closes_1000.loc[:, "Date"] = closes_1000.Date.dt.strftime("%Y-%m-%dT%H:%M:%S%z")
        self.assertListEqual(
            closes_1000.to_dict(orient="records"),
            [{'Date': '2018-04-01T00:00:00', "FI12345": 20.25, "FI23456": 50.59},
             {'Date': '2018-04-02T00:00:00', "FI12345": 20.38, "FI23456": 51.67}]
        )

        volumes = prices.loc["Volume"]
        volumes_930 = volumes.xs("09:30:00", level="Time")
        volumes_930 = volumes_930.reset_index()
        volumes_930.loc[:, "Date"] = volumes_930.Date.dt.strftime("%Y-%m-%dT%H:%M:%S%z")
        self.assertListEqual(
            volumes_930.to_dict(orient="records"),
            [{'Date': '2018-04-01T00:00:00', "FI12345": 1500.0, "FI23456": 9000.0},
             {'Date': '2018-04-02T00:00:00', "FI12345": 1400.0, "FI23456": 1400.0}]
        )

    def test_complain_if_cannot_infer_timezone_from_securities(self):
        """
        Tests error handling when a timezone is not specified and it cannot
        be inferred from the securities master because multiple timezones are
        represented.
        """
        def mock_get_history_db_config(db):
            return {
                "bar_size": "5 mins",
                "universes": ["aapl-arb"],
                "vendor": "ibkr",
                "fields": ["Close","Open","High","Low", "Volume"]
            }

        def mock_download_history_file(code, f, *args, **kwargs):
            prices = pd.DataFrame(
                dict(
                    Sid=[
                        "FI12345",
                        "FI12345",
                        "FI12345",
                        "FI12345",
                        "FI23456",
                        "FI23456",
                        "FI23456",
                        "FI23456"
                        ],
                    Date=[
                        "2018-04-01T09:30:00-04:00",
                        "2018-04-01T10:00:00-04:00",
                        "2018-04-02T09:30:00-04:00",
                        "2018-04-02T10:00:00-04:00",
                        "2018-04-01T08:30:00-05:00",
                        "2018-04-01T09:00:00-05:00",
                        "2018-04-02T08:30:00-05:00",
                        "2018-04-02T09:00:00-05:00",
                        ],
                    Close=[
                        20.10,
                        20.25,
                        20.50,
                        20.38,
                        50.15,
                        50.59,
                        51.59,
                        51.67
                        ],
                    Volume=[
                        1500,
                        7800,
                        1400,
                        800,
                        9000,
                        7100,
                        1400,
                        1500
                    ]
                ))
            prices.to_csv(f, index=False)

        def mock_download_master_file(f, *args, **kwargs):
            securities = pd.DataFrame(dict(Sid=["FI12345","FI23456"],
                                           Timezone=["America/New_York", "America/Mexico_City"]))
            securities.to_csv(f, index=False)
            f.seek(0)

        def mock_list_history_databases():
            return [
                "usa-stk-1d",
                "japan-stk-1d",
                "usa-stk-15min",
                "demo-stk-1min",
                "aapl-arb-5min",
            ]

        def mock_list_realtime_databases():
            return {"demo-stk-taq": ["demo-stk-taq-1h"]}

        def mock_list_bundles():
            return {}

        with patch('quantrocket.price.list_realtime_databases', new=mock_list_realtime_databases):
            with patch('quantrocket.price.list_history_databases', new=mock_list_history_databases):
                with patch('quantrocket.price.get_history_db_config', new=mock_get_history_db_config):
                    with patch('quantrocket.price.download_history_file', new=mock_download_history_file):
                        with patch("quantrocket.price.download_master_file", new=mock_download_master_file):
                            with patch("quantrocket.price.list_bundles", new=mock_list_bundles):
                                with self.assertRaises(ParameterError) as cm:
                                    get_prices(
                                        "aapl-arb-5min",
                                        fields=["Close", "Volume"],
                                    )

        self.assertIn((
            "cannot infer timezone because multiple timezones are present "
            "in data, please specify timezone explicitly (timezones: America/New_York, America/Mexico_City)"
            ), str(cm.exception))

    def test_multiple_timezones(self):
        """
        Tests that multiple timezones are properly handled.
        """
        def mock_get_history_db_config(db):
            return {
                "bar_size": "5 mins",
                "universes": ["aapl-arb"],
                "vendor": "ibkr",
                "fields": ["Close","Open","High","Low", "Volume"]
            }

        def mock_download_history_file(code, f, *args, **kwargs):
            prices = pd.DataFrame(
                dict(
                    Sid=[
                        "FI12345",
                        "FI12345",
                        "FI12345",
                        "FI12345",
                        "FI23456",
                        "FI23456",
                        "FI23456",
                        "FI23456"
                        ],
                    # These bars align but the times are different due to different timezones
                    Date=[
                        "2018-04-01T09:30:00-04:00",
                        "2018-04-01T10:00:00-04:00",
                        "2018-04-02T09:30:00-04:00",
                        "2018-04-02T10:00:00-04:00",
                        "2018-04-01T08:30:00-05:00",
                        "2018-04-01T09:00:00-05:00",
                        "2018-04-02T08:30:00-05:00",
                        "2018-04-02T09:00:00-05:00",
                        ],
                    Close=[
                        20.10,
                        20.25,
                        20.50,
                        20.38,
                        50.15,
                        50.59,
                        51.59,
                        51.67
                        ]
                ))
            prices.to_csv(f, index=False)

        def mock_download_master_file(f, *args, **kwargs):
            securities = pd.DataFrame(dict(Sid=["FI12345","FI23456"],
                                           Timezone=["America/New_York", "America/Mexico_City"]))
            securities.to_csv(f, index=False)
            f.seek(0)

        def mock_list_history_databases():
            return [
                "usa-stk-1d",
                "aapl-arb-5min",
                "japan-stk-1d",
                "usa-stk-15min",
                "demo-stk-1min"
            ]

        def mock_list_realtime_databases():
            return {"demo-stk-taq": ["demo-stk-taq-1h"]}

        def mock_list_bundles():
            return {"usstock-1min": True}

        with patch('quantrocket.price.list_bundles', new=mock_list_bundles):
            with patch('quantrocket.price.list_realtime_databases', new=mock_list_realtime_databases):
                with patch('quantrocket.price.list_history_databases', new=mock_list_history_databases):
                    with patch('quantrocket.price.get_history_db_config', new=mock_get_history_db_config):
                        with patch('quantrocket.price.download_history_file', new=mock_download_history_file):
                            with patch("quantrocket.price.download_master_file", new=mock_download_master_file):

                                prices = get_prices(
                                    "aapl-arb-5min",
                                    fields=["Close"],
                                    timezone="America/New_York"
                                )

        self.assertListEqual(list(prices.index.names), ["Field", "Date", "Time"])
        self.assertEqual(prices.columns.name, "Sid")
        dt = prices.index.get_level_values("Date")
        self.assertTrue(isinstance(dt, pd.DatetimeIndex))
        self.assertIsNone(dt.tz)
        self.assertListEqual(list(prices.columns), ["FI12345","FI23456"])
        self.assertSetEqual(
            set(prices.index.get_level_values("Field")), {"Close"})
        self.assertSetEqual(
            set(prices.index.get_level_values("Time")),
            {"09:30:00", "10:00:00"})
        closes = prices.loc["Close"]
        closes_930 = closes.xs("09:30:00", level="Time")
        closes_930 = closes_930.reset_index()
        closes_930.loc[:, "Date"] = closes_930.Date.dt.strftime("%Y-%m-%dT%H:%M:%S%z")
        self.assertListEqual(
            closes_930.to_dict(orient="records"),
            [{'Date': '2018-04-01T00:00:00', "FI12345": 20.1, "FI23456": 50.15},
             {'Date': '2018-04-02T00:00:00', "FI12345": 20.5, "FI23456": 51.59}]
        )
        closes_1000 = closes.xs("10:00:00", level="Time")
        closes_1000 = closes_1000.reset_index()
        closes_1000.loc[:, "Date"] = closes_1000.Date.dt.strftime("%Y-%m-%dT%H:%M:%S%z")
        self.assertListEqual(
            closes_1000.to_dict(orient="records"),
            [{'Date': '2018-04-01T00:00:00', "FI12345": 20.25, "FI23456": 50.59},
             {'Date': '2018-04-02T00:00:00', "FI12345": 20.38, "FI23456": 51.67}]
        )

        # repeat with a different timezone
        with patch('quantrocket.price.list_bundles', new=mock_list_bundles):
            with patch('quantrocket.price.list_realtime_databases', new=mock_list_realtime_databases):
                with patch('quantrocket.price.list_history_databases', new=mock_list_history_databases):
                    with patch('quantrocket.price.get_history_db_config', new=mock_get_history_db_config):
                        with patch('quantrocket.price.download_history_file', new=mock_download_history_file):
                            with patch("quantrocket.price.download_master_file", new=mock_download_master_file):

                                prices = get_prices(
                                    "aapl-arb-5min",
                                    fields=["Close"],
                                    timezone="America/Mexico_City"
                                )

        self.assertListEqual(list(prices.index.names), ["Field", "Date", "Time"])
        self.assertEqual(prices.columns.name, "Sid")
        dt = prices.index.get_level_values("Date")
        self.assertTrue(isinstance(dt, pd.DatetimeIndex))
        self.assertIsNone(dt.tz)
        self.assertListEqual(list(prices.columns), ["FI12345","FI23456"])
        self.assertSetEqual(
            set(prices.index.get_level_values("Field")), {"Close"})
        self.assertSetEqual(
            set(prices.index.get_level_values("Time")),
            {"08:30:00", "09:00:00"})
        closes = prices.loc["Close"]
        closes_830 = closes.xs("08:30:00", level="Time")
        closes_830 = closes_830.reset_index()
        closes_830.loc[:, "Date"] = closes_830.Date.dt.strftime("%Y-%m-%dT%H:%M:%S%z")
        self.assertListEqual(
            closes_830.to_dict(orient="records"),
            [{'Date': '2018-04-01T00:00:00', "FI12345": 20.1, "FI23456": 50.15},
             {'Date': '2018-04-02T00:00:00', "FI12345": 20.5, "FI23456": 51.59}]
        )
        closes_900 = closes.xs("09:00:00", level="Time")
        closes_900 = closes_900.reset_index()
        closes_900.loc[:, "Date"] = closes_900.Date.dt.strftime("%Y-%m-%dT%H:%M:%S%z")
        self.assertListEqual(
            closes_900.to_dict(orient="records"),
            [{'Date': '2018-04-01T00:00:00', "FI12345": 20.25, "FI23456": 50.59},
             {'Date': '2018-04-02T00:00:00', "FI12345": 20.38, "FI23456": 51.67}]
        )

    def test_intraday_fill_missing_times(self):
        """
        Tests that each date is expanded to include all times represented in
        any date. (This makes the data easier to work when there early close
        days or when the data is queried intraday.)
        """
        def mock_get_history_db_config(db):
            return {
                "bar_size": "2 hours",
                "universes": ["usa-stk"],
                "vendor": "ibkr",
                "fields": ["Close","Open","High","Low", "Volume"]
            }

        def mock_download_history_file(code, f, *args, **kwargs):
            prices = pd.DataFrame(
                dict(
                    Sid=[
                        "FI12345",
                        "FI12345",
                        "FI12345",
                        "FI12345",
                        "FI12345",
                        "FI12345",
                        "FI12345",
                        "FI12345",
                        "FI12345"
                        ],
                    Date=[
                        # complete day
                        "2018-04-01T10:00:00-04:00",
                        "2018-04-01T12:00:00-04:00",
                        "2018-04-01T14:00:00-04:00",
                        # early close
                        "2018-04-02T10:00:00-04:00",
                        "2018-04-02T12:00:00-04:00",
                        # Complete day
                        "2018-04-03T10:00:00-04:00",
                        "2018-04-03T12:00:00-04:00",
                        "2018-04-03T14:00:00-04:00",
                        # intraday query
                        "2018-04-04T10:00:00-04:00",
                        ],
                    Close=[
                        20.10,
                        20.25,
                        20.50,
                        20.38,
                        21.15,
                        22.59,
                        21.59,
                        20.67,
                        21.34
                        ]
                ))
            prices.to_csv(f, index=False)

        def mock_list_history_databases():
            return [
                "usa-stk-1d",
                "japan-stk-1d",
                "usa-stk-2h",
                "usa-stk-15min",
                "demo-stk-1min"
            ]

        def mock_list_realtime_databases():
            return {"demo-stk-taq": ["demo-stk-taq-1h"]}

        def mock_list_bundles():
            return {"usstock-1min": True}

        with patch('quantrocket.price.list_bundles', new=mock_list_bundles):
            with patch('quantrocket.price.list_realtime_databases', new=mock_list_realtime_databases):
                with patch('quantrocket.price.list_history_databases', new=mock_list_history_databases):
                    with patch('quantrocket.price.get_history_db_config', new=mock_get_history_db_config):
                        with patch('quantrocket.price.download_history_file', new=mock_download_history_file):

                            prices = get_prices(
                                "usa-stk-2h",
                                timezone="America/New_York"
                            )

        self.assertListEqual(list(prices.index.names), ["Field", "Date", "Time"])
        self.assertEqual(prices.columns.name, "Sid")
        dt = prices.index.get_level_values("Date")
        self.assertTrue(isinstance(dt, pd.DatetimeIndex))
        self.assertIsNone(dt.tz)
        self.assertListEqual(list(prices.columns), ["FI12345"])
        self.assertSetEqual(
            set(prices.index.get_level_values("Field")),
            {"Close"})
        self.assertSetEqual(
            set(prices.index.get_level_values("Time")),
            {"10:00:00", "12:00:00", "14:00:00"})

        closes = prices.loc["Close"]

        all_days = pd.DatetimeIndex(["2018-04-01", "2018-04-02", "2018-04-03", "2018-04-04"])

        for time in ["10:00:00", "12:00:00", "14:00:00"]:
            idx = closes.xs(time, level="Time").index
            self.assertListEqual(
                list(idx),
                list(all_days),
                msg=time
            )

        # replace nan with "nan" to allow equality comparisons
        closes = closes.where(closes.notnull(), "nan")
        closes = closes["FI12345"]
        self.assertEqual(
            closes.xs("14:00:00", level="Time").loc["2018-04-02"], "nan"
        )
        self.assertEqual(
            closes.xs("12:00:00", level="Time").loc["2018-04-04"], "nan"
        )
        self.assertEqual(
            closes.xs("14:00:00", level="Time").loc["2018-04-04"], "nan"
        )



# Copyright 2017 QuantRocket - All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
from quantrocket.houston import houston
from quantrocket.cli.utils.output import json_to_cli
from quantrocket.cli.utils.files import write_response_to_filepath_or_buffer
from quantrocket.cli.utils.parse import dict_strs_to_dict, dict_to_dict_strs

def download_account_balances(filepath_or_buffer=None, output="csv",
                              start_date=None, end_date=None,
                              latest=False, accounts=None, below=None,
                              fields=None, force_refresh=False):
    """
    Query account balances.

    Parameters
    ----------
    filepath_or_buffer : str or file-like object
        filepath to write the data to, or file-like object (defaults to stdout)

    output : str
        output format (json or csv, default is csv)

    start_date : str (YYYY-MM-DD), optional
        limit to account balance snapshots taken on or after this date

    end_date : str (YYYY-MM-DD), optional
        limit to account balance snapshots taken on or before this date

    latest : bool
        return the latest account balance snapshot

    accounts : list of str, optional
        limit to these accounts

    below : dict of FIELD:AMOUNT, optional
        limit to accounts where the specified field is below the specified
        amount (pass as {field:amount}, for example {'Cushion':0.05})

    fields : list of str, optional
        only return these fields. By default a core set of fields is returned.
        Pass a list of fields, or '*' to return all fields. Pass ['?'] or any
        invalid fieldname to see available fields.

    force_refresh : bool
        refresh account balances to ensure the latest data (default is to query the
        database, which is refreshed every minute)

    Returns
    -------
    None

    Examples
    --------
    Query latest balances. You can use StringIO to load the CSV into pandas.

    >>> f = io.StringIO()
    >>> download_account_balances(f, latest=True)
    >>> balances = pd.read_csv(f, parse_dates=["LastUpdated"])
    """
    params = {}
    if start_date:
        params["start_date"] = start_date
    if end_date:
        params["end_date"] = end_date
    if latest:
        params["latest"] = latest
    if accounts:
        params["accounts"] = accounts
    if below:
        params["below"] = dict_to_dict_strs(below)
    if fields:
        params["fields"] = fields
    if force_refresh:
        params["force_refresh"] = force_refresh

    output = output or "csv"

    if output not in ("csv", "json"):
        raise ValueError("Invalid ouput: {0}".format(output))

    response = houston.get("/account/balances.{0}".format(output), params=params)

    houston.raise_for_status_with_json(response)

    # Don't write a null response to file when using below filters
    if below and response.content[:4] == b"null":
        return

    filepath_or_buffer = filepath_or_buffer or sys.stdout

    write_response_to_filepath_or_buffer(filepath_or_buffer, response)

def _cli_download_account_balances(*args, **kwargs):
    below = kwargs.get("below", None)
    if below:
        kwargs["below"] = dict_strs_to_dict(*below)
    return json_to_cli(download_account_balances, *args, **kwargs)

def download_account_portfolio(filepath_or_buffer=None, output="csv",
                               brokers=None, accounts=None, sec_types=None,
                               exchanges=None, sids=None, symbols=None,
                               include_zero=False, fields=None):
    """
    Download current portfolio.

    Parameters
    ----------
    filepath_or_buffer : str or file-like object
        filepath to write the data to, or file-like object (defaults to stdout)

    output : str
        output format (json or csv, default is csv)

    brokers : list of str, optional
        limit to these brokers. Possible choices: alpaca, ibkr

    accounts : list of str, optional
        limit to these accounts

    sec_types : list of str, optional
        limit to these security types

    exchanges : list of str, optional
        limit to these exchanges

    sids : list of str, optional
        limit to these sids

    symbols : list of str, optional
        limit to these symbols

    include_zero : bool
        include zero position rows (default is to exclude them). Only
        supported for Interactive Brokers.

    fields : list of str, optional
        only return these fields. By default a core set of fields is returned.
        Pass a list of fields, or '*' to return all fields. Pass '?' or any
        invalid fieldname to see available fields.

    Returns
    -------
    None

    Examples
    --------
    Download current portfolio. You can use StringIO to load the CSV into pandas.

    >>> f = io.StringIO()
    >>> download_account_portfolio(f)
    >>> portfolio = pd.read_csv(f, parse_dates=["LastUpdated"])
    """
    params = {}
    if brokers:
        params["brokers"] = brokers
    if accounts:
        params["accounts"] = accounts
    if sec_types:
        params["sec_types"] = sec_types
    if exchanges:
        params["exchanges"] = exchanges
    if sids:
        params["sids"] = sids
    if symbols:
        params["symbols"] = symbols
    if include_zero:
        params["include_zero"] = include_zero
    if fields:
        params["fields"] = fields

    output = output or "csv"

    if output not in ("csv", "json"):
        raise ValueError("Invalid ouput: {0}".format(output))

    response = houston.get("/account/portfolio.{0}".format(output), params=params)

    houston.raise_for_status_with_json(response)

    # Don't write a null response to file
    if response.content[:4] == b"null":
        return

    filepath_or_buffer = filepath_or_buffer or sys.stdout

    write_response_to_filepath_or_buffer(filepath_or_buffer, response)

def _cli_download_account_portfolio(*args, **kwargs):
    return json_to_cli(download_account_portfolio, *args, **kwargs)

def download_exchange_rates(filepath_or_buffer=None, output="csv",
                            start_date=None, end_date=None, latest=False,
                            base_currencies=None, quote_currencies=None):
    """
    Query exchange rates for the base currency.

    The exchange rates in the exchange rate database are sourced from the
    European Central Bank's reference rates, which are updated each day at 4 PM
    CET.

    Parameters
    ----------
    filepath_or_buffer : str or file-like object
        filepath to write the data to, or file-like object (defaults to stdout)

    output : str
        output format (json, csv, default is csv)

    start_date : str (YYYY-MM-DD), optional
        limit to exchange rates on or after this date

    end_date : str (YYYY-MM-DD), optional
        limit to exchange rates on or before this date

    latest : bool
        return the latest exchange rates

    base_currencies : list of str, optional
        limit to these base currencies

    quote_currencies : list of str, optional
        limit to these quote currencies

    Returns
    -------
    None

    Examples
    --------
    Query latest exchange rates. You can use StringIO to load the CSV into pandas.

    >>> f = io.StringIO()
    >>> download_exchange_rates(f, latest=True)
    >>> rates = pd.read_csv(f, parse_dates=["Date"])
    """
    params = {}
    if start_date:
        params["start_date"] = start_date
    if end_date:
        params["end_date"] = end_date
    if latest:
        params["latest"] = latest
    if base_currencies:
        params["base_currencies"] = base_currencies
    if quote_currencies:
        params["quote_currencies"] = quote_currencies

    output = output or "csv"

    if output not in ("csv", "json"):
        raise ValueError("Invalid ouput: {0}".format(output))

    response = houston.get("/account/rates.{0}".format(output), params=params)

    houston.raise_for_status_with_json(response)

    filepath_or_buffer = filepath_or_buffer or sys.stdout

    write_response_to_filepath_or_buffer(filepath_or_buffer, response)

def _cli_download_exchange_rates(*args, **kwargs):
    return json_to_cli(download_exchange_rates, *args, **kwargs)



# Copyright 2017 QuantRocket - All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse

def add_subparser(subparsers):
    _parser = subparsers.add_parser("db", description="QuantRocket database service CLI", help="Backup, restore, and manage databases")
    _subparsers = _parser.add_subparsers(title="subcommands", dest="subcommand")
    _subparsers.required = True

    examples = """
List databases.

Examples:

List all databases:

    quantrocket db list

List all history databases and include details such as file size:

    quantrocket db list --services history --detail

List details for a sharded history database called usa-stk-15min
and list each shard individually:

    quantrocket db list --services history --codes usa-stk-15min --detail --expand
    """
    parser = _subparsers.add_parser(
        "list",
        help="list databases",
        epilog=examples,
        formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.add_argument(
        "-s", "--services",
        nargs="*",
        metavar="SERVICE",
        help="limit to these services")
    parser.add_argument(
        "-c", "--codes",
        nargs="*",
        metavar="DATABASE_CODE",
        help="limit to these codes")
    parser.add_argument(
        "-d", "--detail",
        action="store_true",
        help="return database statistics (default is to return a "
        "flat list of database names)")
    parser.add_argument(
        "-e", "--expand",
        action="store_true",
        help="expand sharded databases to include individual shards "
        "(default is to list sharded databases as a single database)")
    parser.set_defaults(func="quantrocket.db._cli_list_databases")

    examples = """
Set or show Amazon S3 configuration for pushing and pulling databases to and
from S3.

See http://qrok.it/h/dbs3 to learn more.

Credentials are encrypted at rest and never leave your deployment.

Examples:

Configure S3 (will prompt for secret access key):

    quantrocket db s3config --access-key-id XXXXXXXX --bucket my-bucket

Preserve existing credentials but point to a new bucket:

    quantrocket db s3config --bucket my-other-bucket

Show current configuration:

    quantrocket db s3config
    """
    parser = _subparsers.add_parser(
        "s3config",
        help="set or show Amazon S3 configuration for pushing and "
        "pulling databases to and from S3",
        epilog=examples,
        formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.add_argument(
        "-a", "--access-key-id",
        help="AWS access key ID")
    parser.add_argument(
        "-s", "--secret-access-key",
        help="AWS secret access key (if omitted and access-key-id is provided, "
        "will be prompted for secret-access-key)")
    parser.add_argument(
        "-b", "--bucket",
        help="the S3 bucket name to push to/pull from")
    parser.set_defaults(func="quantrocket.db._cli_get_or_set_s3_config")

    examples = """
Push database(s) to Amazon S3.

See http://qrok.it/h/dbs3 to learn more.

Examples:

Push all databases:

    quantrocket db s3push

Push all databases for the history service:

    quantrocket db s3push --services history

Push a database called quantrocket.history.nyse.sqlite:

    quantrocket db s3push --services history --codes nyse
    """
    parser = _subparsers.add_parser(
        "s3push",
        help="push database(s) to Amazon S3",
        epilog=examples,
        formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.add_argument(
        "-s", "--services",
        nargs="*",
        metavar="SERVICE",
        help="limit to these services")
    parser.add_argument(
        "-c", "--codes",
        nargs="*",
        metavar="DATABASE_CODE",
        help="limit to these codes")
    parser.set_defaults(func="quantrocket.db._cli_s3_push_databases")

    examples = """
Pull database(s) from Amazon S3 to the db service.

See http://qrok.it/h/dbs3 to learn more.

Examples:

Pull a database stored on S3 as quantrocket.history.nyse.sqlite.gz:

    quantrocket db s3pull --services history --codes nyse
    """
    parser = _subparsers.add_parser(
        "s3pull",
        help="pull database(s) from Amazon S3",
        epilog=examples,
        formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.add_argument(
        "-s", "--services",
        nargs="*",
        metavar="SERVICE",
        help="limit to these services")
    parser.add_argument(
        "-c", "--codes",
        nargs="*",
        metavar="DATABASE_CODE",
        help="limit to these codes")
    parser.add_argument(
        "-f", "--force",
        action="store_true",
        default=False,
        help="overwrite existing database if one exists (default is to "
        "fail if one exists)")
    parser.set_defaults(func="quantrocket.db._cli_s3_pull_databases")

    examples = """
Optimize databases to improve performance.

This runs the 'VACUUM' command, which defragments the database and
reclaims disk space.

Examples:

Optimize all blotter databases:

    quantrocket db optimize --services blotter
    """
    parser = _subparsers.add_parser(
        "optimize",
        help="optimize databases to improve performance",
        epilog=examples,
        formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.add_argument(
        "-s", "--services",
        nargs="*",
        metavar="SERVICE",
        help="limit to these services")
    parser.add_argument(
        "-c", "--codes",
        nargs="*",
        metavar="DATABASE_CODE",
        help="limit to these codes")
    parser.set_defaults(func="quantrocket.db._cli_optimize_databases")



# Copyright 2017 QuantRocket - All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse

def add_subparser(subparsers):
    _parser = subparsers.add_parser("flightlog", description="QuantRocket logging service CLI", help="Monitor and download logs")
    _subparsers = _parser.add_subparsers(title="subcommands", dest="subcommand")
    _subparsers.required = True

    examples = """
Stream application logs, `tail -f` style.

Examples:

Stream application logs:

    quantrocket flightlog stream

Stream detailed logs:

    quantrocket flightlog stream --detail
    """
    parser = _subparsers.add_parser(
        "stream",
        help="stream application logs, `tail -f` style",
        epilog=examples,
        formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.add_argument(
        "-d", "--detail",
        action="store_true",
        help="show detailed logs from logspout, otherwise show log messages from flightlog only")
    parser.add_argument(
        "--hist",
        type=int,
        metavar="NUM_LINES",
        help="number of log lines to show right away (ignored if showing detailed logs)")
    parser.add_argument(
        "--nocolor",
        action="store_false",
        dest="color",
        help="don't colorize the logs")
    parser.set_defaults(func="quantrocket.flightlog._cli_stream_logs")

    examples = """
Download the logfile.

Examples:

Download application logs:

    quantrocket flightlog get app.log

Download detailed logs:

    quantrocket flightlog get --detail sys.log

Download detailed logs for the history service:

    quantrocket flightlog get --detail --match quantrocket_history sys.log
    """
    parser = _subparsers.add_parser(
        "get",
        help="download the logfile",
        epilog=examples,
        formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.add_argument(
        "outfile",
        metavar="OUTFILE",
        help="filename to write the logfile to")
    parser.add_argument(
        "-d", "--detail",
        action="store_true",
        help="download detailed logs from logspout, otherwise download log messages from "
        "flightlog only")
    parser.add_argument(
        "-m", "--match",
        metavar="PATTERN",
        help="filter the logfile to lines containing this string")
    parser.set_defaults(func="quantrocket.flightlog._cli_download_logfile")

    examples = """
Log a message.

Examples:

Log a message under the name "myapp":

    quantrocket flightlog log "this is a test" --name myapp --level INFO

Log the output from another command:

    quantrocket account balance --below-cushion 0.02 | quantrocket flightlog log --name quantrocket.account --level CRITICAL
    """
    parser = _subparsers.add_parser(
        "log",
        help="log a message",
        epilog=examples,
        formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.add_argument(
        "msg",
        nargs="?",
        default="-",
        help="the message to be logged")
    parser.add_argument(
        "-l", "--level",
        default="INFO",
        metavar="LEVEL",
        choices=("DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"),
        help="the log level for the message. Possible choices: %(choices)s")
    parser.add_argument(
        "-n", "--name",
        dest="logger_name",
        default="quantrocket.cli",
        help="the logger name")
    parser.set_defaults(func="quantrocket.flightlog._cli_log_message")

    examples = """
Set or show the flightlog timezone.

Examples:

Set the flightlog timezone to America/New_York:

    quantrocket flightlog timezone America/New_York

Show the current flightlog timezone:

    quantrocket flightlog timezone
    """
    parser = _subparsers.add_parser(
        "timezone",
        help="set or show the flightlog timezone",
        epilog=examples,
        formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.add_argument(
        "tz",
        nargs="?",
        metavar="TZ",
        help="the timezone to set (pass a partial timezone string such as 'newyork' "
        "or 'europe' to see close matches, or pass '?' to see all choices)")
    parser.set_defaults(func="quantrocket.flightlog._cli_get_or_set_timezone")

    examples = """
Set or show the Papertrail log configuration.

See http://qrok.it/h/pt to learn more.

Examples:

Set the Papertrail host and port to log to:

    quantrocket flightlog papertrail --host logs.papertrailapp.com --port 55555

Show the current papertrail config:

    quantrocket flightlog papertrail
    """
    parser = _subparsers.add_parser(
        "papertrail",
        help="set or show the Papertrail log configuration",
        epilog=examples,
        formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.add_argument(
        "--host",
        metavar="HOST",
        help="the Papertrail host to log to")
    parser.add_argument(
        "--port",
        metavar="PORT",
        type=int,
        help="the Papertrail port to log to")
    parser.set_defaults(func="quantrocket.flightlog._cli_get_or_set_papertrail_config")



# Copyright 2017 QuantRocket - All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import six
import sys

def write_response_to_filepath_or_buffer(filepath_or_buffer, response):
    """
    Writes the response content to the filepath or buffer.
    """
    if hasattr(filepath_or_buffer, "write"):
        if six.PY3 and filepath_or_buffer is sys.stdout:
            # Write bytes to stdout (https://stackoverflow.com/a/23932488)
            filepath_or_buffer = filepath_or_buffer.buffer
        mode = getattr(filepath_or_buffer, "mode", "w")
        for chunk in response.iter_content(chunk_size=1024):
            if chunk:
                if "b" not in mode and six.PY3:
                    chunk = chunk.decode("utf-8")
                filepath_or_buffer.write(chunk)
        if filepath_or_buffer.seekable():
            filepath_or_buffer.seek(0)
    else:
        with open(filepath_or_buffer, "wb") as f:
            for chunk in response.iter_content(chunk_size=1024):
                if chunk:
                    f.write(chunk)



#!/usr/bin/env python
#
# Copyright 2017 QuantRocket - All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging, logging.handlers
import socket
import six
import sys
import os
from six.moves import queue, urllib
from .exceptions import ImproperlyConfigured
from .houston import Houston, houston
from quantrocket.cli.utils.output import json_to_cli

FLIGHTLOG_PATH = "/flightlog/handler"

LOG_RECORD_TYPE_BOUNDARY = "||||q5%XfK4#||||"

class _ImpatientHttpHandler(logging.handlers.HTTPHandler):
    """
    An HttpHandler that sets a short timeout (instead of the default no
    timeout), as well as serializing the record attribute types so they can
    be reconstituted.
    """

    def mapLogRecord(self, record):
        dict_with_types = {}
        for k, v in record.__dict__.items():
            type_name = type(v).__name__
            dict_with_types[k] = "{0}{1}{2}".format(
                type_name, LOG_RECORD_TYPE_BOUNDARY, v)
        return dict_with_types

    def emit(self, record):
        orig_timeout = socket.getdefaulttimeout()
        socket.setdefaulttimeout(3)
        super(_ImpatientHttpHandler, self).emit(record)
        socket.setdefaulttimeout(orig_timeout)

def FlightlogHandler(background=None):
    """
    Returns a log handler that logs to flightlog.

    Parameters
    ----------
    background : bool
        If True, causes logging to happen in a background thread so that logging
        doesn't block. Background logging requires Python 3.2 or higher,
        and defaults to True for supported versions and False otherwise.

    Returns
    -------
    `logging.handlers.QueueHandler` or `quantrocket.flightlog._ImpatientHttpHandler`

    Examples
    --------
    Log a message using the FlightlogHandler:

    >>> import logging
    >>> from quantrocket.flightlog import FlightlogHandler
    >>> logger = logging.getLogger('myapp')
    >>> logger.setLevel(logging.DEBUG)
    >>> handler = FlightlogHandler()
    >>> logger.addHandler(handler)
    >>> logger.info('my app just opened a position')
    """
    base_url = os.environ.get("HOUSTON_URL", None)
    if not base_url:
        raise ImproperlyConfigured("HOUSTON_URL is not set")
    parsed = urllib.parse.urlparse(base_url)
    secure = parsed.scheme == "https"
    if "HOUSTON_USERNAME" in os.environ and "HOUSTON_PASSWORD" in os.environ:
        credentials = (os.environ["HOUSTON_USERNAME"], os.environ["HOUSTON_PASSWORD"])
    else:
        credentials = None

    path = os.environ.get("FLIGHTLOG_PATH") or FLIGHTLOG_PATH

    if six.PY2:
        if secure:
            raise NotImplementedError("Logging to Flightlog over HTTPS requires Python 3")
        if credentials:
            raise NotImplementedError("Logging to Flightlog using Basic Auth requires Python 3")

    if six.PY2:
        http_handler = _ImpatientHttpHandler(
            parsed.netloc, path, method="POST")
    else:
        http_handler = _ImpatientHttpHandler(
            parsed.netloc, path, method="POST", secure=secure, credentials=credentials)

    if six.PY2 or sys.version_info.minor < 2:
        if background:
            import warnings
            warnings.warn('Background logging requires Python 3.2 or higher. Logging in foreground...')
        background = False
    elif background is None:
        background = True

    if background:
        log_queue = queue.Queue(-1)  # no limit on size
        queue_handler = logging.handlers.QueueHandler(log_queue)
        listener = logging.handlers.QueueListener(log_queue, http_handler)
        listener.start()
        return queue_handler
    else:
        return http_handler

def _cli_log_message(msg, logger_name=None, level="INFO"):
    """
    Log a single message to Flightlog. Intended for CLI usage. Calling this
    function multiple times within the same process will configure duplicate
    handlers and result in duplicate messages.
    """
    logger = logging.getLogger(logger_name)
    levelnum = logging.getLevelName(level.upper())
    try:
        int(levelnum)
    except ValueError:
        raise ValueError("level must be one of DEBUG, INFO, WARNING, ERROR, CRITICAL")

    handler = FlightlogHandler(background=False)
    logger.addHandler(handler)
    logger.setLevel(levelnum)
    if msg == "-":
        msg = sys.stdin.read()
    for line in msg.splitlines():
        if line:
            logger.log(levelnum, line)

    exit_code = 0
    return None, exit_code

def stream_logs(detail=False, hist=None, color=True):
    """
    Stream application logs, `tail -f` style.

    Parameters
    ----------
    detail : bool
        if True, show detailed logs from logspout, otherwise show log messages
        from flightlog only (default False)

    hist : int, optional
        number of log lines to show right away (ignored if showing detailed logs)

    color : bool
        colorize the logs

    Yields
    -------
    str
        each log line as it arrives
    """
    params = {}
    if detail:
        path = "/logspout/logs"
        if not color:
            params["colors"] = "off"
    else:
        path = "/flightlog/stream/logs"
        if hist:
            params['hist'] = hist
        if not color:
            params["nocolor"] = "true"

    houston = Houston()
    response = houston.get(path, stream=True, params=params)
    houston.raise_for_status_with_json(response)
    try:
        for line in response.iter_lines():
            if six.PY3:
                line = line.decode("utf-8")
            yield line
    except KeyboardInterrupt:
        houston.close()
        return

def _cli_print_stream(*args, **kwargs):
    generator = stream_logs(*args, **kwargs)
    for chunk in generator:
        try:
            # disable output buffering using flush to allow grepping
            # stream (flush arg not available before Python 3.3);
            # called via exec in order to catch SyntaxError on Python 2
            exec("print(chunk, flush=True)")
        except (SyntaxError, TypeError):
            print(chunk)

def _cli_stream_logs(*args, **kwargs):
    return json_to_cli(_cli_print_stream, *args, **kwargs)

def download_logfile(outfile, detail=False, match=None):
    """
    Download the logfile.

    Parameters
    ----------
    outfile: str, required
        filename to write the logfile to

    detail : bool
        if True, show detailed logs from logspout, otherwise show log messages
        from flightlog only (default False)

    match : str, optional
        filter the logfile to lines containing this string

    Returns
    -------
    None
    """
    if detail:
        logtype = "system"
    else:
        logtype = "app"

    params = {}
    if match:
        params["match"] = match

    response = houston.get("/flightlog/logfile/{0}".format(logtype), params=params, stream=True)
    houston.raise_for_status_with_json(response)

    if response.status_code == 204:
        return response.json()

    with open(outfile, "wb") as f:
        for chunk in response.iter_content(chunk_size=1024):
            if chunk:
                f.write(chunk)

def _cli_download_logfile(*args, **kwargs):
    return json_to_cli(download_logfile, *args, **kwargs)

def get_timezone():
    """
    Return the flightlog timezone.

    Returns
    -------
    dict
        dict with key timezone
    """
    response = houston.get("/flightlog/timezone")
    houston.raise_for_status_with_json(response)
    return response.json()

def set_timezone(tz):
    """
    Set the flightlog timezone.

    Parameters
    ----------
    tz : str, required
        the timezone to set (pass a partial timezone string such as 'newyork'
        or 'europe' to see close matches, or pass '?' to see all choices)

    Returns
    -------
    dict
        status message

    Examples
    --------
    Set the flightlog timezone to America/New_York:

    >>> set_timezone("America/New_York")
    """
    params = {"tz": tz}
    response = houston.put("/flightlog/timezone", params=params)
    houston.raise_for_status_with_json(response)
    return response.json()

def _cli_get_or_set_timezone(tz=None, *args, **kwargs):
    if tz:
        return json_to_cli(set_timezone, tz, *args, **kwargs)
    else:
        return json_to_cli(get_timezone, *args, **kwargs)

def get_papertrail_config():
    """
    Return the current Papertrail log configuration, if any.

    See http://qrok.it/h/pt to learn more.

    Returns
    -------
    dict
        config details
    """
    response = houston.get("/flightlog/papertrail")
    houston.raise_for_status_with_json(response)
    # It's possible to get a 204 empty response
    if not response.content:
        return {}
    return response.json()

def set_papertrail_config(host, port):
    """
    Set the Papertrail log configuration.

    See http://qrok.it/h/pt to learn more.

    Parameters
    ----------
    host : str, required
        the Papertrail host to log to

    port : int, required
        the Papertrail port to log to

    Returns
    -------
    dict
        status message

    Examples
    --------
    Configure flightlog to log to Papertrail:

    >>> set_papertrail_config("logs.papertrailapp.com", 55555)
    """
    params = {"host": host, "port": port}
    response = houston.put("/flightlog/papertrail", params=params)
    houston.raise_for_status_with_json(response)
    return response.json()

def _cli_get_or_set_papertrail_config(host=None, port=None, *args, **kwargs):
    if host or port:
        return json_to_cli(set_papertrail_config, host, port, *args, **kwargs)
    else:
        return json_to_cli(get_papertrail_config, *args, **kwargs)
